#!/usr/bin/env python3
'''
t3_fs_api.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import errno
import logging
import os
import shutil
import stat
import tempfile
from argparse import Namespace
from random import randint

import pytest
import trio
from common import CLOCK_GRANULARITY, safe_sleep
from pyfuse3 import FUSEError
from pytest import raises as assert_raises
from pytest_checklogs import assert_logs
from t2_block_cache import DummyQueue

from s3ql import ROOT_INODE, fs
from s3ql.backends import local
from s3ql.backends.comprenc import ComprencBackend
from s3ql.backends.pool import BackendPool
from s3ql.block_cache import BlockCache
from s3ql.database import Connection, FsAttributes, create_tables
from s3ql.fsck import Fsck
from s3ql.inode_cache import CACHE_SIZE as INODE_CACHE_SIZE
from s3ql.inode_cache import InodeCache
from s3ql.mkfs import init_tables

# We need to access to protected members
# pylint: disable=W0212


# The classes provided by pyfuse3 have read-only attributes,
# so we duck-type our own.
class Ctx:
    def __init__(self):
        self.uid = randint(0, 2**32)
        self.gid = randint(0, 2**32)
        self.pid = randint(0, 2**32)
        self.umask = 0


class SetattrFields:
    def __init__(self, **kw):
        self.update_atime = False
        self.update_mtime = False
        self.update_mode = False
        self.update_uid = False
        self.update_gid = False
        self.update_size = False
        self.__dict__.update(kw)


some_ctx = Ctx()


@pytest.fixture
async def ctx():
    ctx = Namespace()
    ctx.backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')
    plain_backend = local.Backend(Namespace(storage_url='local://' + ctx.backend_dir))
    ctx.backend_pool = BackendPool(lambda: ComprencBackend(b'schwubl', ('zlib', 6), plain_backend))
    ctx.backend = ctx.backend_pool.pop_conn()
    ctx.cachedir = tempfile.mkdtemp(prefix='s3ql-cache-')
    ctx.max_obj_size = 1024

    # Destructors are not guaranteed to run, and we can't unlink
    # the file immediately because apsw refers to it by name.
    # Therefore, we unlink the file manually in tearDown()
    ctx.dbfile = tempfile.NamedTemporaryFile(delete=False)

    ctx.db = Connection(ctx.dbfile.name)
    create_tables(ctx.db)
    init_tables(ctx.db)

    cache = BlockCache(ctx.backend_pool, ctx.db, ctx.cachedir + "/cache", ctx.max_obj_size * 5)
    cache.trio_token = trio.lowlevel.current_trio_token()
    ctx.cache = cache
    ctx.server = fs.Operations(cache, ctx.db, ctx.max_obj_size, InodeCache(ctx.db, 0))
    ctx.server.init()

    # Monkeypatch around the need for removal and upload threads
    cache.to_remove = DummyQueue(cache)

    class DummyChannel:
        async def send(self, arg):
            await trio.to_thread.run_sync(cache._do_upload, *arg)

    cache.to_upload = (DummyChannel(), None)

    # Keep track of unused filenames
    ctx.name_cnt = 0

    yield ctx

    ctx.server.inodes.destroy()
    await ctx.cache.destroy()
    ctx.db.close()
    os.unlink(ctx.dbfile.name)
    shutil.rmtree(ctx.cachedir)
    shutil.rmtree(ctx.backend_dir)


def random_data(len_):
    with open("/dev/urandom", "rb") as fd:
        return fd.read(len_)


async def fsck(ctx):
    await ctx.cache.drop()
    ctx.server.inodes.flush()
    fsck = Fsck(
        ctx.cachedir + '/cache',
        ctx.backend,
        FsAttributes(
            data_block_size=ctx.max_obj_size,
            metadata_block_size=ctx.max_obj_size,
        ),
        ctx.db,
    )
    fsck.check()
    assert not fsck.found_errors


def newname(ctx):
    ctx.name_cnt += 1
    return ("s3ql_%d" % ctx.name_cnt).encode()


async def test_getattr_root(ctx):
    assert stat.S_ISDIR((await ctx.server.getattr(ROOT_INODE, some_ctx)).st_mode)
    await fsck(ctx)


async def test_create(ctx):
    fctx = Ctx()
    mode = dir_mode()
    name = newname(ctx)

    inode_p_old = await ctx.server.getattr(ROOT_INODE, some_ctx)
    safe_sleep(CLOCK_GRANULARITY)
    ctx.server._create(ROOT_INODE, name, mode, fctx)

    id_ = ctx.db.get_val(
        'SELECT inode FROM contents JOIN names ON name_id = names.id '
        'WHERE name=? AND parent_inode = ?',
        (name, ROOT_INODE),
    )

    inode = await ctx.server.getattr(id_, some_ctx)

    assert inode.st_mode == mode
    assert inode.st_uid == fctx.uid
    assert inode.st_gid == fctx.gid
    assert inode.st_nlink == 1
    assert inode.st_size == 0

    inode_p_new = await ctx.server.getattr(ROOT_INODE, some_ctx)

    assert inode_p_new.st_mtime_ns > inode_p_old.st_mtime_ns
    assert inode_p_new.st_ctime_ns > inode_p_old.st_ctime_ns

    await ctx.server.forget([(id_, 1)])
    await fsck(ctx)


async def test_extstat(ctx):
    # Test with zero contents
    ctx.server.extstat()

    # Test with empty file
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.release(fh)
    ctx.server.extstat()

    # Test with data in file
    fi = await ctx.server.open(inode.st_ino, os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, b'foobar')
    await ctx.server.release(fh)

    ctx.server.extstat()
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


def dir_mode():
    return (randint(0, 0o7777) & ~stat.S_IFDIR) | stat.S_IFDIR


def file_mode():
    return (randint(0, 0o7777) & ~stat.S_IFREG) | stat.S_IFREG


async def test_getxattr(ctx):
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.release(fh)

    with assert_raises(FUSEError):
        await ctx.server.getxattr(inode.st_ino, b'nonexistent-attr', some_ctx)

    await ctx.server.setxattr(inode.st_ino, b'my-attr', b'strabumm!', some_ctx)
    assert await ctx.server.getxattr(inode.st_ino, b'my-attr', some_ctx) == b'strabumm!'

    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_link(ctx):
    name = newname(ctx)

    inode_p_new = await ctx.server.mkdir(ROOT_INODE, newname(ctx), dir_mode(), some_ctx)
    inode_p_new_before = await ctx.server.getattr(inode_p_new.st_ino, some_ctx)

    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.release(fh)
    safe_sleep(CLOCK_GRANULARITY)

    inode_before = await ctx.server.getattr(inode.st_ino, some_ctx)
    await ctx.server.link(inode.st_ino, inode_p_new.st_ino, name, some_ctx)

    inode_after = await ctx.server.lookup(inode_p_new.st_ino, name, some_ctx)
    inode_p_new_after = await ctx.server.getattr(inode_p_new.st_ino, some_ctx)

    id_ = ctx.db.get_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (name, inode_p_new.st_ino),
    )

    assert inode_before.st_ino == id_
    assert inode_after.st_nlink == 2
    assert inode_after.st_ctime_ns > inode_before.st_ctime_ns
    assert inode_p_new_before.st_mtime_ns < inode_p_new_after.st_mtime_ns
    assert inode_p_new_before.st_ctime_ns < inode_p_new_after.st_ctime_ns
    await ctx.server.forget([(inode.st_ino, 1), (inode_p_new.st_ino, 1), (inode_after.st_ino, 1)])
    await fsck(ctx)


async def test_listxattr(ctx):
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.release(fh)

    assert await ctx.server.listxattr(inode.st_ino, some_ctx) == []

    await ctx.server.setxattr(inode.st_ino, b'key1', b'blub', some_ctx)
    assert [b'key1'] == await ctx.server.listxattr(inode.st_ino, some_ctx)

    await ctx.server.setxattr(inode.st_ino, b'key2', b'blub', some_ctx)
    assert sorted([b'key1', b'key2']) == sorted(await ctx.server.listxattr(inode.st_ino, some_ctx))
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_read(ctx):
    len_ = ctx.max_obj_size
    data = random_data(len_)
    off = ctx.max_obj_size // 2
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh

    await ctx.server.write(fh, off, data)
    inode_before = await ctx.server.getattr(inode.st_ino, some_ctx)
    safe_sleep(CLOCK_GRANULARITY)
    assert await ctx.server.read(fh, off, len_) == data
    inode_after = await ctx.server.getattr(inode.st_ino, some_ctx)
    assert inode_after.st_atime_ns > inode_before.st_atime_ns
    assert await ctx.server.read(fh, 0, len_) == b"\0" * off + data[:off]
    assert await ctx.server.read(fh, ctx.max_obj_size, len_) == data[off:]
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_readdir(ctx, monkeypatch):
    # Create a few entries
    names = [('entry_%2d' % i).encode() for i in range(20)]
    for name in names:
        (fi, inode) = await ctx.server.create(ROOT_INODE, name, file_mode(), os.O_RDWR, some_ctx)
        fh = fi.fh
        await ctx.server.release(fh)
        await ctx.server.forget([(inode.st_ino, 1)])

    # Delete some to make sure that we don't have continuous rowids
    remove_no = [0, 2, 3, 5, 9]
    for i in remove_no:
        await ctx.server.unlink(ROOT_INODE, names[i], some_ctx)
        del names[i]

    # Read all
    fh = await ctx.server.opendir(ROOT_INODE, some_ctx)
    names_ret = []
    next_ids = []

    def readdir_reply(token, name, attr, next_id):
        names_ret.append(name)
        next_ids.append(next_id)
        return True

    monkeypatch.setattr('pyfuse3.readdir_reply', readdir_reply)
    await ctx.server.readdir(fh, 0, 'my-token')
    await ctx.server.releasedir(fh)
    assert set(names + [b'lost+found']) == set(names_ret)

    # Read in parts
    fh = await ctx.server.opendir(ROOT_INODE, some_ctx)
    names_ret.clear()
    next_ids.clear()
    while True:
        old_len = len(next_ids)
        await ctx.server.readdir(fh, next_ids[-1] if next_ids else 0, 'my-token')
        if old_len == len(next_ids):
            break

    assert set(names + [b'lost+found']) == set(names_ret)
    await ctx.server.releasedir(fh)

    await fsck(ctx)


async def test_forget(ctx):
    name = newname(ctx)

    # Test that entries are deleted when they're no longer referenced
    (fi, inode) = await ctx.server.create(ROOT_INODE, name, file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, b'foobar')
    await ctx.server.unlink(ROOT_INODE, name, some_ctx)
    assert not ctx.db.has_val(
        'SELECT 1 FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (name, ROOT_INODE),
    )
    assert (await ctx.server.getattr(inode.st_ino, some_ctx)).st_ino
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])

    assert not ctx.db.has_val('SELECT 1 FROM inodes WHERE id=?', (inode.st_ino,))

    await fsck(ctx)


async def test_removexattr(ctx):
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.release(fh)

    with assert_raises(FUSEError):
        await ctx.server.removexattr(inode.st_ino, b'some name', some_ctx)

    await ctx.server.setxattr(inode.st_ino, b'key1', b'blub', some_ctx)
    await ctx.server.removexattr(inode.st_ino, b'key1', some_ctx)
    assert await ctx.server.listxattr(inode.st_ino, some_ctx) == []
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_rename(ctx):
    oldname = newname(ctx)
    newname_ = newname(ctx)

    inode = await ctx.server.mkdir(ROOT_INODE, oldname, dir_mode(), some_ctx)

    inode_p_new = await ctx.server.mkdir(ROOT_INODE, newname(ctx), dir_mode(), some_ctx)
    inode_p_new_before = await ctx.server.getattr(inode_p_new.st_ino, some_ctx)
    inode_p_old_before = await ctx.server.getattr(ROOT_INODE, some_ctx)
    safe_sleep(CLOCK_GRANULARITY)

    await ctx.server.rename(ROOT_INODE, oldname, inode_p_new.st_ino, newname_, 0, some_ctx)

    inode_p_old_after = await ctx.server.getattr(ROOT_INODE, some_ctx)
    inode_p_new_after = await ctx.server.getattr(inode_p_new.st_ino, some_ctx)

    assert not ctx.db.has_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (oldname, ROOT_INODE),
    )
    id_ = ctx.db.get_val(
        'SELECT inode FROM contents JOIN names ON names.id == name_id '
        'WHERE name=? AND parent_inode = ?',
        (newname_, inode_p_new.st_ino),
    )
    assert inode.st_ino == id_

    assert inode_p_new_before.st_mtime_ns < inode_p_new_after.st_mtime_ns
    assert inode_p_new_before.st_ctime_ns < inode_p_new_after.st_ctime_ns
    assert inode_p_old_before.st_mtime_ns < inode_p_old_after.st_mtime_ns
    assert inode_p_old_before.st_ctime_ns < inode_p_old_after.st_ctime_ns

    await ctx.server.forget([(inode.st_ino, 1), (inode_p_new.st_ino, 1)])
    await fsck(ctx)


async def test_rename2(ctx):
    src_file_name = newname(ctx)
    src_folder_name = newname(ctx)
    dst_file_name = newname(ctx)
    dst_folder_name = newname(ctx)

    inode_dst_folder = await ctx.server.mkdir(ROOT_INODE, dst_folder_name, dir_mode(), some_ctx)
    inode_src_folder = await ctx.server.mkdir(ROOT_INODE, src_folder_name, dir_mode(), some_ctx)
    await ctx.server.create(
        inode_src_folder.st_ino, src_file_name, file_mode(), os.O_RDWR, some_ctx
    )

    safe_sleep(CLOCK_GRANULARITY)

    # Create files in order to flush the inode cache of the last 3 inodes
    for _ in range(0, INODE_CACHE_SIZE):
        await ctx.server.create(ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx)

    # Load the inodes in a specific order
    fi = await ctx.server.open(ROOT_INODE, os.O_RDWR, some_ctx)
    await ctx.server.release(fi.fh)
    fi = await ctx.server.open(inode_dst_folder.st_ino, os.O_RDWR, some_ctx)
    await ctx.server.release(fi.fh)
    fi = await ctx.server.open(inode_src_folder.st_ino, os.O_RDWR, some_ctx)
    await ctx.server.release(fi.fh)

    # Put these inodes on the brink of being flushed in the inode cache
    for _ in range(0, INODE_CACHE_SIZE - 3):
        await ctx.server.create(ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx)

    # Move the file
    await ctx.server.rename(
        inode_src_folder.st_ino, src_file_name, inode_dst_folder.st_ino, dst_file_name, 0, some_ctx
    )

    await fsck(ctx)


async def test_replace_file(ctx):
    oldname = newname(ctx)
    newname_ = newname(ctx)

    (fi, inode) = await ctx.server.create(ROOT_INODE, oldname, file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, b'some data to deal with')
    await ctx.server.release(fh)
    await ctx.server.setxattr(inode.st_ino, b'test_xattr', b'42*8', some_ctx)

    inode_p_new = await ctx.server.mkdir(ROOT_INODE, newname(ctx), dir_mode(), some_ctx)
    inode_p_new_before = await ctx.server.getattr(inode_p_new.st_ino, some_ctx)
    inode_p_old_before = await ctx.server.getattr(ROOT_INODE, some_ctx)

    (fi, inode2) = await ctx.server.create(
        inode_p_new.st_ino, newname_, file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, b'even more data to deal with')
    await ctx.server.release(fh)
    await ctx.server.setxattr(inode2.st_ino, b'test_xattr', b'42*8', some_ctx)
    await ctx.server.forget([(inode2.st_ino, 1)])

    safe_sleep(CLOCK_GRANULARITY)
    await ctx.server.rename(ROOT_INODE, oldname, inode_p_new.st_ino, newname_, 0, some_ctx)

    inode_p_old_after = await ctx.server.getattr(ROOT_INODE, some_ctx)
    inode_p_new_after = await ctx.server.getattr(inode_p_new.st_ino, some_ctx)

    assert not ctx.db.has_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (oldname, ROOT_INODE),
    )
    id_ = ctx.db.get_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (newname_, inode_p_new.st_ino),
    )
    assert inode.st_ino == id_

    assert inode_p_new_before.st_mtime_ns < inode_p_new_after.st_mtime_ns
    assert inode_p_new_before.st_ctime_ns < inode_p_new_after.st_ctime_ns
    assert inode_p_old_before.st_mtime_ns < inode_p_old_after.st_mtime_ns
    assert inode_p_old_before.st_ctime_ns < inode_p_old_after.st_ctime_ns

    assert not ctx.db.has_val('SELECT id FROM inodes WHERE id=?', (inode2.st_ino,))
    await ctx.server.forget([(inode.st_ino, 1), (inode_p_new.st_ino, 1)])
    await fsck(ctx)


async def test_replace_dir(ctx):
    oldname = newname(ctx)
    newname_ = newname(ctx)

    inode = await ctx.server.mkdir(ROOT_INODE, oldname, dir_mode(), some_ctx)

    inode_p_new = await ctx.server.mkdir(ROOT_INODE, newname(ctx), dir_mode(), some_ctx)
    inode_p_new_before = await ctx.server.getattr(inode_p_new.st_ino, some_ctx)
    inode_p_old_before = await ctx.server.getattr(ROOT_INODE, some_ctx)

    inode2 = await ctx.server.mkdir(inode_p_new.st_ino, newname_, dir_mode(), some_ctx)
    await ctx.server.forget([(inode2.st_ino, 1)])

    safe_sleep(CLOCK_GRANULARITY)
    await ctx.server.rename(ROOT_INODE, oldname, inode_p_new.st_ino, newname_, 0, some_ctx)

    inode_p_old_after = await ctx.server.getattr(ROOT_INODE, some_ctx)
    inode_p_new_after = await ctx.server.getattr(inode_p_new.st_ino, some_ctx)

    assert not ctx.db.has_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (oldname, ROOT_INODE),
    )
    id_ = ctx.db.get_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (newname_, inode_p_new.st_ino),
    )
    assert inode.st_ino == id_

    assert inode_p_new_before.st_mtime_ns < inode_p_new_after.st_mtime_ns
    assert inode_p_new_before.st_ctime_ns < inode_p_new_after.st_ctime_ns
    assert inode_p_old_before.st_mtime_ns < inode_p_old_after.st_mtime_ns
    assert inode_p_old_before.st_ctime_ns < inode_p_old_after.st_ctime_ns

    await ctx.server.forget([(inode.st_ino, 1), (inode_p_new.st_ino, 1)])
    assert not ctx.db.has_val('SELECT id FROM inodes WHERE id=?', (inode2.st_ino,))
    await fsck(ctx)


async def test_setattr_one(ctx):
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.release(fh)
    inode_old = await ctx.server.getattr(inode.st_ino, some_ctx)

    attr = await ctx.server.getattr(inode.st_ino, some_ctx)  # this is a fresh instance
    attr.st_mode = file_mode()
    attr.st_uid = randint(0, 2**32)
    attr.st_gid = randint(0, 2**32)  # should be ignored
    attr.st_atime_ns = randint(0, 2**50)
    attr.st_mtime_ns = randint(0, 2**50)

    safe_sleep(CLOCK_GRANULARITY)
    sf = SetattrFields(update_mode=True, update_uid=True, update_atime=True, update_mtime=True)
    await ctx.server.setattr(inode.st_ino, attr, sf, None, some_ctx)
    inode_new = await ctx.server.getattr(inode.st_ino, some_ctx)

    for name in ('st_mode', 'st_uid', 'st_atime_ns', 'st_mtime_ns'):
        assert getattr(attr, name) == getattr(inode_new, name)
    for name in ('st_gid', 'st_size', 'st_nlink', 'st_rdev', 'st_blocks', 'st_blksize'):
        assert getattr(inode_old, name) == getattr(inode_new, name)
    assert inode_old.st_ctime_ns < inode_new.st_ctime_ns

    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_setattr_two(ctx):
    (fi, inode_old) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh

    attr = await ctx.server.getattr(inode_old.st_ino, some_ctx)
    attr.st_mode = file_mode()
    attr.st_uid = randint(0, 2**32)
    attr.st_gid = randint(0, 2**32)
    attr.st_mtime_ns = randint(0, 2**50)
    attr.st_ctime_ns = 5e9

    safe_sleep(CLOCK_GRANULARITY)
    sf = SetattrFields(update_gid=True, update_mtime=True)

    await ctx.server.setattr(inode_old.st_ino, attr, sf, None, some_ctx)
    inode_new = await ctx.server.getattr(inode_old.st_ino, some_ctx)

    for name in ('st_gid', 'st_mtime_ns'):
        assert getattr(attr, name) == getattr(inode_new, name)
    for name in (
        'st_uid',
        'st_size',
        'st_nlink',
        'st_rdev',
        'st_blocks',
        'st_blksize',
        'st_mode',
        'st_atime_ns',
    ):
        assert getattr(inode_old, name) == getattr(inode_new, name)
    assert inode_old.st_ctime_ns < inode_new.st_ctime_ns

    await ctx.server.release(fh)
    await ctx.server.forget([(inode_old.st_ino, 1)])
    await fsck(ctx)


async def test_setattr_large(ctx):
    (fi, inode_old) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh

    attr = await ctx.server.getattr(inode_old.st_ino, some_ctx)
    attr.st_mtime_ns = int(2**63 + 1)
    sf = SetattrFields(update_mtime=True)

    with pytest.raises(FUSEError) as exc:
        await ctx.server.setattr(inode_old.st_ino, attr, sf, None, some_ctx)
    assert exc.value.errno == errno.EINVAL

    await ctx.server.release(fh)
    await ctx.server.forget([(inode_old.st_ino, 1)])
    await fsck(ctx)


async def test_truncate(ctx):
    len_ = int(2.7 * ctx.max_obj_size)
    data = random_data(len_)

    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, data)

    attr = await ctx.server.getattr(inode.st_ino, some_ctx)
    attr.st_size = len_ // 2
    sf = SetattrFields(update_size=True)
    await ctx.server.setattr(inode.st_ino, attr, sf, None, some_ctx)
    assert await ctx.server.read(fh, 0, len_) == data[: len_ // 2]
    attr.st_size = len_
    await ctx.server.setattr(inode.st_ino, attr, sf, None, some_ctx)
    assert await ctx.server.read(fh, 0, len_) == data[: len_ // 2] + b'\0' * (len_ // 2)
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_truncate_0(ctx):
    len1 = 158
    len2 = 133

    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, random_data(len1))
    await ctx.server.release(fh)
    ctx.server.inodes.flush()

    attr = await ctx.server.getattr(inode.st_ino, some_ctx)
    fi = await ctx.server.open(inode.st_ino, os.O_RDWR, some_ctx)
    fh = fi.fh
    attr.st_size = 0
    await ctx.server.setattr(inode.st_ino, attr, SetattrFields(update_size=True), fh, some_ctx)
    data = random_data(len2)
    await ctx.server.write(fh, 0, data)
    await ctx.server.release(fh)
    ctx.server.inodes.flush()
    fi = await ctx.server.open(inode.st_ino, os.O_RDONLY, some_ctx)
    fh = fi.fh
    assert await ctx.server.read(fh, 0, len2 + 1) == data
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_setxattr(ctx):
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.release(fh)

    await ctx.server.setxattr(inode.st_ino, b'my-attr', b'strabumm!', some_ctx)
    assert await ctx.server.getxattr(inode.st_ino, b'my-attr', some_ctx) == b'strabumm!'
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_names(ctx):
    name1 = newname(ctx)
    name2 = newname(ctx)

    (fi, inode) = await ctx.server.create(ROOT_INODE, name1, file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])

    (fi, inode) = await ctx.server.create(ROOT_INODE, name2, file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.release(fh)

    await ctx.server.setxattr(inode.st_ino, name1, b'strabumm!', some_ctx)
    await fsck(ctx)

    await ctx.server.removexattr(inode.st_ino, name1, some_ctx)
    await fsck(ctx)

    await ctx.server.setxattr(inode.st_ino, name1, b'strabumm karacho!!', some_ctx)
    await ctx.server.unlink(ROOT_INODE, name1, some_ctx)
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_statfs(ctx):
    # Test with zero contents
    await ctx.server.statfs(some_ctx)

    # Test with empty file
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.release(fh)
    await ctx.server.statfs(some_ctx)

    # Test with data in file
    fi = await ctx.server.open(inode.st_ino, os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, b'foobar')
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])
    await ctx.server.statfs(some_ctx)


async def test_symlink(ctx):
    target = newname(ctx)
    name = newname(ctx)

    inode_p_before = await ctx.server.getattr(ROOT_INODE, some_ctx)
    safe_sleep(CLOCK_GRANULARITY)
    inode = await ctx.server.symlink(ROOT_INODE, name, target, some_ctx)
    inode_p_after = await ctx.server.getattr(ROOT_INODE, some_ctx)

    assert target, await ctx.server.readlink(inode.st_ino == some_ctx)

    id_ = ctx.db.get_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (name, ROOT_INODE),
    )

    assert inode.st_ino == id_
    assert inode_p_before.st_mtime_ns < inode_p_after.st_mtime_ns
    assert inode_p_before.st_ctime_ns < inode_p_after.st_ctime_ns

    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_unlink(ctx):
    name = newname(ctx)

    (fi, inode) = await ctx.server.create(ROOT_INODE, name, file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, b'some data to deal with')
    await ctx.server.release(fh)

    # Add extended attributes
    await ctx.server.setxattr(inode.st_ino, b'test_xattr', b'42*8', some_ctx)
    await ctx.server.forget([(inode.st_ino, 1)])

    inode_p_before = await ctx.server.getattr(ROOT_INODE, some_ctx)
    safe_sleep(CLOCK_GRANULARITY)
    await ctx.server.unlink(ROOT_INODE, name, some_ctx)
    inode_p_after = await ctx.server.getattr(ROOT_INODE, some_ctx)

    assert inode_p_before.st_mtime_ns < inode_p_after.st_mtime_ns
    assert inode_p_before.st_ctime_ns < inode_p_after.st_ctime_ns

    assert not ctx.db.has_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (name, ROOT_INODE),
    )
    assert not ctx.db.has_val('SELECT id FROM inodes WHERE id=?', (inode.st_ino,))

    await fsck(ctx)


async def test_rmdir(ctx):
    name = newname(ctx)
    inode = await ctx.server.mkdir(ROOT_INODE, name, dir_mode(), some_ctx)
    await ctx.server.forget([(inode.st_ino, 1)])
    inode_p_before = await ctx.server.getattr(ROOT_INODE, some_ctx)
    safe_sleep(CLOCK_GRANULARITY)
    await ctx.server.rmdir(ROOT_INODE, name, some_ctx)
    inode_p_after = await ctx.server.getattr(ROOT_INODE, some_ctx)

    assert inode_p_before.st_mtime_ns < inode_p_after.st_mtime_ns
    assert inode_p_before.st_ctime_ns < inode_p_after.st_ctime_ns
    assert not ctx.db.has_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (name, ROOT_INODE),
    )
    assert not ctx.db.has_val('SELECT id FROM inodes WHERE id=?', (inode.st_ino,))

    await fsck(ctx)


async def test_relink(ctx):
    name = newname(ctx)
    name2 = newname(ctx)
    data = b'some data to deal with'

    (fi, inode) = await ctx.server.create(ROOT_INODE, name, file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, data)
    await ctx.server.release(fh)
    await ctx.server.unlink(ROOT_INODE, name, some_ctx)
    ctx.server.inodes.flush()
    assert not ctx.db.has_val(
        'SELECT inode FROM contents JOIN names ON names.id = name_id '
        'WHERE name=? AND parent_inode = ?',
        (name, ROOT_INODE),
    )
    assert ctx.db.has_val('SELECT id FROM inodes WHERE id=?', (inode.st_ino,))
    await ctx.server.link(inode.st_ino, ROOT_INODE, name2, some_ctx)
    await ctx.server.forget([(inode.st_ino, 2)])

    inode = await ctx.server.lookup(ROOT_INODE, name2, some_ctx)
    fi = await ctx.server.open(inode.st_ino, os.O_RDONLY, some_ctx)
    fh = fi.fh
    assert await ctx.server.read(fh, 0, len(data)) == data
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_write(ctx):
    len_ = ctx.max_obj_size
    data = random_data(len_)
    off = ctx.max_obj_size // 2
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    inode_before = await ctx.server.getattr(inode.st_ino, some_ctx)
    safe_sleep(CLOCK_GRANULARITY)
    await ctx.server.write(fh, off, data)
    inode_after = await ctx.server.getattr(inode.st_ino, some_ctx)

    assert inode_after.st_mtime_ns > inode_before.st_mtime_ns
    assert inode_after.st_ctime_ns > inode_before.st_ctime_ns
    assert inode_after.st_size == off + len_

    await ctx.server.write(fh, 0, data)
    inode_after = await ctx.server.getattr(inode.st_ino, some_ctx)
    assert inode_after.st_size == off + len_

    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_fsync(ctx):
    len_ = ctx.max_obj_size
    data = random_data(len_)
    off = ctx.max_obj_size // 2
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, off, data)
    await ctx.server.fsync(fh, datasync=False)
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])


async def test_failsafe(ctx):
    len_ = ctx.max_obj_size
    data = random_data(len_)
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, data)
    await ctx.cache.drop()
    assert not ctx.server.failsafe

    datafile = os.path.join(ctx.backend_dir, 's3ql_data_', 's3ql_data_1')
    shutil.copy(datafile, datafile + '.bak')

    # Modify contents
    with open(datafile, 'rb+') as rfh:
        rfh.seek(560)
        rfh.write(b'blrub!')
    with assert_raises(FUSEError) as cm:
        with assert_logs('^Backend returned malformed data for', count=1, level=logging.ERROR):
            await ctx.server.read(fh, 0, len_)
    assert cm.value.errno == errno.EIO
    assert ctx.server.failsafe

    # Restore contents, but should be marked as damaged now
    os.rename(datafile + '.bak', datafile)
    with assert_raises(FUSEError) as cm:
        await ctx.server.read(fh, 0, len_)
    assert cm.value.errno == errno.EIO

    # Release and re-open, now we should be able to access again
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])

    # ..but not write access since we are in failsafe mode
    with assert_raises(FUSEError) as cm:
        await ctx.server.open(inode.st_ino, os.O_RDWR, some_ctx)
    assert cm.value.errno == errno.EPERM

    # ..ready only is fine.
    fi = await ctx.server.open(inode.st_ino, os.O_RDONLY, some_ctx)
    fh = fi.fh
    await ctx.server.read(fh, 0, len_)

    # Remove completely, should give error after cache flush
    os.unlink(datafile)
    await ctx.server.read(fh, 3, len_ // 2)
    await ctx.cache.drop()
    with assert_raises(FUSEError) as cm:
        with assert_logs('^Backend lost block', count=1, level=logging.ERROR):
            await ctx.server.read(fh, 5, len_ // 2)
    assert cm.value.errno == errno.EIO

    # Don't call fsck, we're missing a block


async def test_create_open(ctx):
    name = newname(ctx)
    # Create a new file
    (fi, inode) = await ctx.server.create(ROOT_INODE, name, file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])

    # Open it atomically
    (fi, inode) = await ctx.server.create(ROOT_INODE, name, file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])

    await fsck(ctx)


async def test_open_truncate(ctx):
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, random_data(int(1.5 * ctx.max_obj_size)))
    await ctx.server.release(fh)
    ctx.server.inodes.flush()

    data = b'this data overwrites the previous data'
    fi = await ctx.server.open(inode.st_ino, os.O_RDWR | os.O_TRUNC, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, data)
    await ctx.server.release(fh)
    ctx.server.inodes.flush()

    fi = await ctx.server.open(inode.st_ino, os.O_RDONLY, some_ctx)
    fh = fi.fh
    assert await ctx.server.read(fh, 0, len(data) + 1) == data
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_edit(ctx):
    len_ = ctx.max_obj_size
    data = random_data(len_)
    (fi, inode) = await ctx.server.create(
        ROOT_INODE, newname(ctx), file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, data)
    await ctx.server.release(fh)

    await ctx.cache.drop()

    fi = await ctx.server.open(inode.st_ino, os.O_RDWR, some_ctx)
    fh = fi.fh
    attr = await ctx.server.getattr(inode.st_ino, some_ctx)
    attr.st_size = 0
    await ctx.server.setattr(inode.st_ino, attr, SetattrFields(update_size=True), fh, some_ctx)
    await ctx.server.write(fh, 0, data[50:])
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])
    await fsck(ctx)


async def test_copy_tree(ctx, monkeypatch):
    ext_attr_name = b'system.foo.brazl'
    ext_attr_val = b'schulla dku woumm bramp'

    src_inode = await ctx.server.mkdir(ROOT_INODE, b'source', dir_mode(), some_ctx)
    dst_inode = await ctx.server.mkdir(ROOT_INODE, b'dest', dir_mode(), some_ctx)

    # Create file
    (fi, f1_inode) = await ctx.server.create(
        src_inode.st_ino, b'file1', file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, b'file1 contents')
    await ctx.server.release(fh)
    await ctx.server.setxattr(f1_inode.st_ino, ext_attr_name, ext_attr_val, some_ctx)

    # Create hardlink
    (fi, f2_inode) = await ctx.server.create(
        src_inode.st_ino, b'file2', file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, b'file2 contents')
    await ctx.server.release(fh)
    f2_inode = await ctx.server.link(f2_inode.st_ino, src_inode.st_ino, b'file2_hardlink', some_ctx)

    # Create subdirectory
    d1_inode = await ctx.server.mkdir(src_inode.st_ino, b'dir1', dir_mode(), some_ctx)
    d2_inode = await ctx.server.mkdir(d1_inode.st_ino, b'dir2', dir_mode(), some_ctx)

    # ..with a 3rd hardlink
    f2_inode = await ctx.server.link(f2_inode.st_ino, d1_inode.st_ino, b'file2_hardlink', some_ctx)

    # Replicate
    monkeypatch.setattr('pyfuse3.invalidate_inode', lambda *args: None)
    await ctx.server.copy_tree(src_inode.st_ino, dst_inode.st_ino)

    # Change files
    fi = await ctx.server.open(f1_inode.st_ino, os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, b'new file1 contents')
    await ctx.server.release(fh)

    fi = await ctx.server.open(f2_inode.st_ino, os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.write(fh, 0, b'new file2 contents')
    await ctx.server.release(fh)

    # Get copy properties
    f1_inode_c = await ctx.server.lookup(dst_inode.st_ino, b'file1', some_ctx)
    f2_inode_c = await ctx.server.lookup(dst_inode.st_ino, b'file2', some_ctx)
    f2h_inode_c = await ctx.server.lookup(dst_inode.st_ino, b'file2_hardlink', some_ctx)
    d1_inode_c = await ctx.server.lookup(dst_inode.st_ino, b'dir1', some_ctx)
    d2_inode_c = await ctx.server.lookup(d1_inode_c.st_ino, b'dir2', some_ctx)
    f2_h_inode_c = await ctx.server.lookup(d1_inode_c.st_ino, b'file2_hardlink', some_ctx)

    # Check file1
    fi = await ctx.server.open(f1_inode_c.st_ino, os.O_RDWR, some_ctx)
    fh = fi.fh
    assert await ctx.server.read(fh, 0, 42) == b'file1 contents'
    await ctx.server.release(fh)
    assert f1_inode.st_ino != f1_inode_c.st_ino
    assert await ctx.server.getxattr(f1_inode_c.st_ino, ext_attr_name, some_ctx) == ext_attr_val

    # Check file2
    fi = await ctx.server.open(f2_inode_c.st_ino, os.O_RDWR, some_ctx)
    fh = fi.fh
    assert await ctx.server.read(fh, 0, 42) == b'file2 contents'
    await ctx.server.release(fh)
    assert f2_inode_c.st_ino == f2h_inode_c.st_ino
    assert f2_inode_c.st_nlink == 3
    assert f2_inode.st_ino != f2_inode_c.st_ino
    assert f2_h_inode_c.st_ino == f2_inode_c.st_ino

    # Check subdir1
    assert d1_inode.st_ino != d1_inode_c.st_ino
    assert d2_inode.st_ino != d2_inode_c.st_ino
    await ctx.server.forget(list(ctx.server.open_inodes.items()))
    await fsck(ctx)


async def test_copy_tree_2(ctx, monkeypatch):
    src_inode = await ctx.server.mkdir(ROOT_INODE, b'source', dir_mode(), some_ctx)
    dst_inode = await ctx.server.mkdir(ROOT_INODE, b'dest', dir_mode(), some_ctx)

    # Create file
    (fi, inode) = await ctx.server.create(
        src_inode.st_ino, b'file1', file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, b'block 1 contents')
    await ctx.server.write(fh, ctx.max_obj_size, b'block 1 contents')
    await ctx.server.release(fh)
    await ctx.server.forget([(inode.st_ino, 1)])

    monkeypatch.setattr('pyfuse3.invalidate_inode', lambda *args: None)
    await ctx.server.copy_tree(src_inode.st_ino, dst_inode.st_ino)

    await ctx.server.forget([(src_inode.st_ino, 1), (dst_inode.st_ino, 1)])
    await fsck(ctx)


async def test_lock_tree(ctx):
    inode1 = await ctx.server.mkdir(ROOT_INODE, b'source', dir_mode(), some_ctx)

    # Create file
    (fi, inode1a) = await ctx.server.create(
        inode1.st_ino, b'file1', file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, b'file1 contents')
    await ctx.server.release(fh)

    # Create subdirectory
    inode2 = await ctx.server.mkdir(inode1.st_ino, b'dir1', dir_mode(), some_ctx)
    (fi, inode2a) = await ctx.server.create(
        inode2.st_ino, b'file2', file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, b'file2 contents')
    await ctx.server.release(fh)

    # Another file
    (fi, inode3) = await ctx.server.create(ROOT_INODE, b'file1', file_mode(), os.O_RDWR, some_ctx)
    fh = fi.fh
    await ctx.server.release(fh)

    # Lock
    await ctx.server.lock_tree(inode1.st_ino)

    for i in (inode1.st_ino, inode1a.st_ino, inode2.st_ino, inode2a.st_ino):
        assert ctx.server.inodes[i].locked

    # Remove
    with assert_raises(FUSEError) as cm:
        await ctx.server._remove(inode1.st_ino, b'file1', inode1a.st_ino)
    assert cm.value.errno == errno.EPERM

    # Rename / Replace
    with assert_raises(FUSEError) as cm:
        await ctx.server.rename(ROOT_INODE, b'file1', inode1.st_ino, b'file2', 0, some_ctx)
    assert cm.value.errno == errno.EPERM
    with assert_raises(FUSEError) as cm:
        await ctx.server.rename(inode1.st_ino, b'file1', ROOT_INODE, b'file2', 0, some_ctx)
    assert cm.value.errno == errno.EPERM

    # Open
    with assert_raises(FUSEError) as cm:
        await ctx.server.open(inode2a.st_ino, os.O_RDWR, some_ctx)
    assert cm.value.errno == errno.EPERM
    with assert_raises(FUSEError) as cm:
        await ctx.server.open(inode2a.st_ino, os.O_WRONLY, some_ctx)
    assert cm.value.errno == errno.EPERM
    await ctx.server.release(await ctx.server.open(inode3.st_ino, os.O_WRONLY, some_ctx))

    # Write
    fi = await ctx.server.open(inode2a.st_ino, os.O_RDONLY, some_ctx)
    fh = fi.fh
    with assert_raises(FUSEError) as cm:
        await ctx.server.write(fh, 0, b'foo')
    assert cm.value.errno == errno.EPERM
    await ctx.server.release(fh)

    # Create
    with assert_raises(FUSEError) as cm:
        ctx.server._create(inode2.st_ino, b'dir1', dir_mode(), os.O_RDWR, some_ctx)
    assert cm.value.errno == errno.EPERM

    # Setattr
    attr = await ctx.server.getattr(inode2a.st_ino, some_ctx)
    with assert_raises(FUSEError) as cm:
        await ctx.server.setattr(
            inode2a.st_ino, attr, SetattrFields(update_mtime=True), None, some_ctx
        )
    assert cm.value.errno == errno.EPERM

    # xattr
    with assert_raises(FUSEError) as cm:
        await ctx.server.setxattr(inode2.st_ino, b'name', b'value', some_ctx)
    assert cm.value.errno == errno.EPERM
    with assert_raises(FUSEError) as cm:
        await ctx.server.removexattr(inode2.st_ino, b'name', some_ctx)
    assert cm.value.errno == errno.EPERM
    await ctx.server.forget(list(ctx.server.open_inodes.items()))
    await fsck(ctx)


async def test_remove_tree(ctx, monkeypatch):
    inode1 = await ctx.server.mkdir(ROOT_INODE, b'source', dir_mode(), some_ctx)

    # Create file
    (fi, inode1a) = await ctx.server.create(
        inode1.st_ino, b'file1', file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, b'file1 contents')
    await ctx.server.release(fh)

    # Create subdirectory
    inode2 = await ctx.server.mkdir(inode1.st_ino, b'dir1', dir_mode(), some_ctx)
    (fi, inode2a) = await ctx.server.create(
        inode2.st_ino, b'file2', file_mode(), os.O_RDWR, some_ctx
    )
    fh = fi.fh
    await ctx.server.write(fh, 0, b'file2 contents')
    await ctx.server.release(fh)

    # Remove
    await ctx.server.forget(list(ctx.server.open_inodes.items()))
    monkeypatch.setattr('pyfuse3.invalidate_entry', lambda *args: None)
    await ctx.server.remove_tree(ROOT_INODE, b'source')

    for id_p, name in (
        (ROOT_INODE, b'source'),
        (inode1.st_ino, b'file1'),
        (inode1.st_ino, b'dir1'),
        (inode2.st_ino, b'file2'),
    ):
        assert not ctx.db.has_val(
            'SELECT inode FROM contents JOIN names ON names.id = name_id '
            'WHERE name=? AND parent_inode = ?',
            (name, id_p),
        )

    for id_ in (inode1.st_ino, inode1a.st_ino, inode2.st_ino, inode2a.st_ino):
        assert not ctx.db.has_val('SELECT id FROM inodes WHERE id=?', (id_,))

    await fsck(ctx)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
setup.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

try:
    import setuptools
except ModuleNotFoundError:
    raise SystemExit(
        'Setuptools package not found. Please install from '
        'https://pypi.python.org/pypi/setuptools'
    )
import faulthandler
import os
import subprocess
import sys
from glob import glob

from setuptools import Extension
from setuptools.command.test import test as TestCommand

faulthandler.enable()

basedir = os.path.abspath(os.path.dirname(sys.argv[0]))
DEVELOPER_MODE = os.path.exists(os.path.join(basedir, 'MANIFEST.in'))
if DEVELOPER_MODE:
    print('MANIFEST.in exists, running in developer mode')

# Add S3QL sources
sys.path.insert(0, os.path.join(basedir, 'src'))
sys.path.insert(0, os.path.join(basedir, 'util'))
import s3ql


class pytest(TestCommand):
    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(['tests'])
        sys.exit(errno)


def main():
    with open(os.path.join(basedir, 'README.rst'), 'r') as fh:
        long_desc = fh.read()

    compile_args = ['-Wall', '-Wextra', '-Wconversion', '-Wsign-compare']

    # Enable all fatal warnings only when compiling from Mercurial tip.
    # (otherwise we break forward compatibility because compilation with newer
    # compiler may fail if additional warnings are added)
    if DEVELOPER_MODE:
        if os.environ.get('CI') != 'true':
            compile_args.append('-Werror')

        # Value-changing conversions should always be explicit.
        compile_args.append('-Werror=conversion')

        # Note that (i > -1) is false if i is unsigned (-1 will be converted to
        # a large positive value). We certainly don't want to do this by
        # accident.
        compile_args.append('-Werror=sign-compare')

        # These warnings have always been harmless, and have always been due to
        # issues in Cython code rather than S3QL. Cython itself warns if there
        # are unused variables in .pyx code.
        compile_args.append('-Wno-unused-parameter')
        compile_args.append('-Wno-unused-function')

    required_pkgs = [
        'apsw >= 3.42.0',  # https://github.com/rogerbinns/apsw/issues/459
        'cryptography',
        'requests',
        'defusedxml',
        'google-auth',
        'google-auth-oauthlib',
        # Need trio.lowlevel
        'trio >= 0.15',
        'pyfuse3 >= 3.2.0, < 4.0',
    ]

    setuptools.setup(
        name='s3ql',
        zip_safe=False,
        version=s3ql.VERSION,
        description='a full-featured file system for online data storage',
        long_description=long_desc,
        author='Nikolaus Rath',
        author_email='Nikolaus@rath.org',
        license='GPLv3',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: No Input/Output (Daemon)',
            'Environment :: Console',
            'License :: OSI Approved :: GNU Library or Lesser General Public License (GPLv3)',
            'Topic :: Internet',
            'Operating System :: POSIX',
            'Topic :: System :: Archiving',
        ],
        platforms=['POSIX', 'UNIX', 'Linux'],
        package_dir={'': 'src'},
        packages=setuptools.find_packages('src'),
        provides=['s3ql'],
        ext_modules=[
            Extension(
                's3ql.sqlite3ext',
                ['src/s3ql/sqlite3ext.cpp'],
                extra_compile_args=compile_args,
                language='c++',
                depends=['src/s3ql/_sqlite3ext.cpp'],
            ),
        ],
        data_files=[
            (
                'share/man/man1',
                [
                    os.path.join('doc/manpages/', x)
                    for x in glob(os.path.join(basedir, 'doc', 'manpages', '*.1'))
                ],
            )
        ],
        entry_points={
            'console_scripts': [
                'mkfs.s3ql = s3ql.mkfs:main',
                'fsck.s3ql = s3ql.fsck:main',
                'mount.s3ql = s3ql.mount:main',
                'umount.s3ql = s3ql.umount:main',
                's3qlcp = s3ql.cp:main',
                's3qlstat = s3ql.statfs:main',
                's3qladm = s3ql.adm:main',
                's3qlctrl = s3ql.ctrl:main',
                's3qllock = s3ql.lock:main',
                's3qlrm = s3ql.remove:main',
                's3ql_oauth_client = s3ql.oauth_client:main',
                's3ql_verify = s3ql.verify:main',
            ]
        },
        install_requires=required_pkgs,
        tests_require=['pytest >= 3.7'],
        python_requires='>=3.8',
        cmdclass={'upload_docs': upload_docs, 'build_cython': build_cython, 'pytest': pytest},
    )


class build_cython(setuptools.Command):
    user_options = []
    boolean_options = []
    description = "Compile .pyx to .c/.cpp"

    def initialize_options(self):
        pass

    def finalize_options(self):
        # Attribute defined outside init
        # pylint: disable=W0201
        self.extensions = self.distribution.ext_modules

    def run(self):
        try:
            from Cython.Build import cythonize
        except ImportError:
            raise SystemExit('Cython needs to be installed for this command') from None

        for extension in self.extensions:
            for fpath in extension.sources:
                (file_, ext) = os.path.splitext(fpath)
                spath = os.path.join(basedir, file_ + '.pyx')
                if ext not in ('.c', '.cpp') or not os.path.exists(spath):
                    continue
                print('compiling %s to %s' % (file_ + '.pyx', file_ + ext))
                cythonize(spath, language_level=3)


class upload_docs(setuptools.Command):
    user_options = []
    boolean_options = []
    description = "Upload documentation"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.check_call(
            [
                'rsync',
                '-aHv',
                '--del',
                os.path.join(basedir, 'doc', 'html') + '/',
                'ebox.rath.org:/srv/www.rath.org/s3ql-docs/',
            ]
        )
        subprocess.check_call(
            [
                'rsync',
                '-aHv',
                '--del',
                os.path.join(basedir, 'doc', 'manual.pdf'),
                'ebox.rath.org:/srv/www.rath.org/s3ql-docs/',
            ]
        )


if __name__ == '__main__':
    main()




# distutils: language = c++
# cython: language_level=3

from libcpp.unordered_map cimport unordered_map
from libcpp.unordered_set cimport unordered_set
from libcpp.string cimport string
from cython.operator cimport dereference as deref

import os


cdef extern from "_sqlite3ext.cpp":
    ctypedef int size_t
    ctypedef unordered_set[size_t] block_map_t
    cdef unordered_map[string, block_map_t] file_block_map
    cdef size_t blocksize
    cdef string vfsname


cdef class WriteTracker:
    cdef block_map_t* block_map

    # Constructors can only accept Python objects, so we need a factory method to construct from a
    # C++ object.
    # https://cython.readthedocs.io/en/stable/src/userguide/extension_types.html#existing-pointers-instantiation
    @staticmethod
    cdef WriteTracker make(block_map_t* block_map):
        # Call to __new__ bypasses __init__ constructor
        cdef WriteTracker tracker = WriteTracker.__new__(WriteTracker)
        tracker.block_map = block_map
        return tracker

    def get_block(self):
        it = self.block_map.begin()
        if it == self.block_map.end():
            raise KeyError()
        cdef size_t blockno = deref(it)
        self.block_map.erase(it)
        return blockno

    def get_count(self):
        return self.block_map.size()

    def clear(self):
        self.block_map.clear()

def track_writes(fpath: str):
    cdef string path_c = os.fsencode(fpath)
    return WriteTracker.make(&file_block_map[path_c])

def set_blocksize(size: int):
    global blocksize
    if blocksize != 0 and blocksize != size:
        raise ValueError('block size may only be set once')
    blocksize = size

def reset():
    global blocksize
    blocksize = 0
    file_block_map.clear()

def get_vfsname():
    return vfsname.decode('utf-8')


#!/usr/bin/env python3
'''
t3_fsck.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import _thread
import hashlib
import os
import shutil
import stat
import tempfile
import unittest
from argparse import Namespace

from s3ql import ROOT_INODE
from s3ql.backends import local
from s3ql.common import time_ns
from s3ql.database import Connection, FsAttributes, NoSuchRowError, create_tables
from s3ql.fsck import Fsck
from s3ql.mkfs import init_tables


def sha256(s):
    return hashlib.sha256(s).digest()


class fsck_tests(unittest.TestCase):
    def setUp(self):
        self.backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')
        self.backend = local.Backend(Namespace(storage_url='local://' + self.backend_dir))
        self.cachedir = tempfile.mkdtemp(prefix='s3ql-cache-')
        self.max_obj_size = 1024

        self.dbfile = tempfile.NamedTemporaryFile()
        self.db = Connection(self.dbfile.name)
        create_tables(self.db)
        init_tables(self.db)

        self.fsck = Fsck(
            self.cachedir,
            self.backend,
            FsAttributes(data_block_size=self.max_obj_size, metadata_block_size=self.max_obj_size),
            self.db,
        )

    def tearDown(self):
        shutil.rmtree(self.cachedir)
        shutil.rmtree(self.backend_dir)
        self.dbfile.close()

    def assert_fsck(self, *fns, can_fix=True):
        '''Check that fns detects and corrects errors'''

        self.fsck.expect_errors = True
        for fn in fns:
            self.fsck.found_errors = False
            fn()
            assert self.fsck.found_errors
        if not can_fix:
            return
        self.fsck.found_errors = False
        self.fsck.expect_errors = False
        self.fsck.check()
        assert not self.fsck.found_errors

    def test_cache(self):
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG
                | stat.S_IRUSR
                | stat.S_IWUSR
                | stat.S_IXUSR
                | stat.S_IRGRP
                | stat.S_IXGRP
                | stat.S_IROTH
                | stat.S_IXOTH,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                8,
            ),
        )
        self._link(b'test-entry', inode)

        # Create new block
        with open(self.cachedir + '/%d-0' % inode, 'wb') as fh:
            fh.write(b'somedata')
        size = os.stat(fh.name).st_size
        self.assert_fsck(self.fsck.check_cache)
        self.assertEqual(self.backend['s3ql_data_1'], b'somedata')
        assert self.db.get_val('SELECT size FROM inodes WHERE id=?', (inode,)) == size

        # Existing block
        self.db.execute('UPDATE inodes SET size=? WHERE id=?', (self.max_obj_size + 8, inode))
        with open(self.cachedir + '/%d-1' % inode, 'wb') as fh:
            fh.write(b'somedata')
        self.assert_fsck(self.fsck.check_cache)

        # Old block preserved
        with open(self.cachedir + '/%d-0' % inode, 'wb') as fh:
            fh.write(b'somedat2')
        self.assert_fsck(self.fsck.check_cache)

        # Old block removed
        with open(self.cachedir + '/%d-1' % inode, 'wb') as fh:
            fh.write(b'somedat3')
        self.assert_fsck(self.fsck.check_cache)

    def test_lof1(self):
        # Make lost+found a file
        inode = self.db.get_val(
            "SELECT inode FROM contents_v WHERE name=? AND parent_inode=?",
            (b"lost+found", ROOT_INODE),
        )
        self.db.execute('DELETE FROM contents WHERE parent_inode=?', (inode,))
        self.db.execute(
            'UPDATE inodes SET mode=?, size=? WHERE id=?',
            (stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR, 0, inode),
        )

        def check():
            self.fsck.check_lof()
            self.fsck.check_inodes_refcount()

        self.assert_fsck(check)

    def test_lof2(self):
        # Remove lost+found
        name_id = self.db.get_val('SELECT id FROM names WHERE name=?', (b'lost+found',))
        inode = self.db.get_val(
            'SELECT inode FROM contents WHERE name_id=? AND parent_inode=?',
            (name_id, ROOT_INODE),
        )
        self.db.execute('DELETE FROM inodes WHERE id=?', (inode,))
        self.db.execute(
            'DELETE FROM contents WHERE name_id=? and parent_inode=?', (name_id, ROOT_INODE)
        )
        self.db.execute('UPDATE names SET refcount = refcount-1 WHERE id=?', (name_id,))

        self.assert_fsck(self.fsck.check_lof)

    def test_lof_issue_273(self):
        name_id = self.db.get_val('SELECT id FROM names WHERE name=?', (b'lost+found',))
        inode = self.db.get_val(
            'SELECT inode FROM contents WHERE name_id=? AND parent_inode=?',
            (name_id, ROOT_INODE),
        )
        self.db.execute('UPDATE contents SET parent_inode = ? WHERE inode = ?', (inode, inode))

        self.assert_fsck(self.fsck.check_lof)

    def test_wrong_inode_refcount(self):
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                0,
            ),
        )
        self._link(b'name1', inode)
        self._link(b'name2', inode)
        self.assert_fsck(self.fsck.check_inodes_refcount)

    def test_orphaned_inode(self):
        self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                0,
            ),
        )
        self.assert_fsck(self.fsck.check_inodes_refcount)

    def test_name_refcount(self):
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                2,
                0,
            ),
        )
        self._link(b'name1', inode)
        self._link(b'name2', inode)

        self.db.execute('UPDATE names SET refcount=refcount+1 WHERE name=?', (b'name1',))

        self.assert_fsck(self.fsck.check_names_refcount)

    def test_orphaned_name(self):
        self._add_name(b'zupbrazl')
        self.assert_fsck(self.fsck.check_names_refcount)

    def test_contents_inode(self):
        self.db.execute(
            'INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)',
            (self._add_name(b'foobar'), 124, ROOT_INODE),
        )

        self.assert_fsck(self.fsck.check_contents_inode)

    def test_contents_inode_p(self):
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                0,
            ),
        )
        self.db.execute(
            'INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)',
            (self._add_name(b'foobar'), inode, 123),
        )

        self.assert_fsck(self.fsck.check_contents_parent_inode)

    def test_contents_name(self):
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                0,
            ),
        )
        self.db.execute(
            'INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)',
            (42, inode, ROOT_INODE),
        )

        self.assert_fsck(self.fsck.check_contents_name)

    def _add_name(self, name):
        '''Get id for *name* and increase refcount

        Name is inserted in table if it does not yet exist.
        '''

        try:
            name_id = self.db.get_val('SELECT id FROM names WHERE name=?', (name,))
        except NoSuchRowError:
            name_id = self.db.rowid('INSERT INTO names (name, refcount) VALUES(?,?)', (name, 1))
        else:
            self.db.execute('UPDATE names SET refcount=refcount+1 WHERE id=?', (name_id,))
        return name_id

    def _link(self, name, inode, parent_inode=ROOT_INODE):
        '''Link /*name* to *inode*'''

        self.db.execute(
            'INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)',
            (self._add_name(name), inode, parent_inode),
        )

    def test_inodes_size(self):
        id_ = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                128,
            ),
        )
        self._link(b'test-entry', id_)

        block_size = self.max_obj_size // 3
        obj_id = self.db.rowid(
            'INSERT INTO objects (refcount, phys_size, length, hash) VALUES(?, ?, ?, ?)',
            (1, 36, block_size, sha256(b'foo')),
        )
        self.backend['s3ql_data_%d' % obj_id] = b'foo'

        # One block, no holes, size plausible
        self.db.execute('UPDATE inodes SET size=? WHERE id=?', (block_size, id_))
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?, ?, ?)', (id_, 0, obj_id)
        )
        self.fsck.found_errors = False
        self.fsck.check()
        assert not self.fsck.found_errors

        # One block, size not plausible
        self.db.execute('UPDATE inodes SET size=? WHERE id=?', (block_size - 1, id_))
        self.assert_fsck(self.fsck.check_inodes_size)

        # Two blocks, hole at the beginning, size plausible
        self.db.execute('DELETE FROM inode_blocks WHERE inode=?', (id_,))
        self.db.execute(
            'UPDATE inodes SET size=? WHERE id=?', (self.max_obj_size + block_size, id_)
        )
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?, ?, ?)', (id_, 1, obj_id)
        )
        self.fsck.found_errors = False
        self.fsck.check()
        assert not self.fsck.found_errors

        # Two blocks, no holes, size plausible
        self.db.execute('UPDATE objects SET refcount = 2 WHERE id = ?', (obj_id,))
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?, ?, ?)', (id_, 0, obj_id)
        )
        self.fsck.found_errors = False
        self.fsck.check()
        assert not self.fsck.found_errors

        # Two blocks, size not plausible
        self.db.execute(
            'UPDATE inodes SET size=? WHERE id=?', (self.max_obj_size + block_size - 1, id_)
        )
        self.assert_fsck(self.fsck.check_inodes_size)

        # Two blocks, hole at the end, size plausible
        self.db.execute(
            'UPDATE inodes SET size=? WHERE id=?', (self.max_obj_size + block_size + 1, id_)
        )
        self.fsck.found_errors = False
        self.fsck.check()
        assert not self.fsck.found_errors

        # Two blocks, size not plausible
        self.db.execute('UPDATE inodes SET size=? WHERE id=?', (self.max_obj_size, id_))
        self.assert_fsck(self.fsck.check_inodes_size)

    def test_objects_id(self):
        # Create an object that only exists in the backend
        self.backend['s3ql_data_4364'] = b'Testdata'
        self.assert_fsck(self.fsck.check_objects_id)

        # Create an object that does not exist in the backend
        self.db.execute(
            'INSERT INTO objects (id, refcount, phys_size, length) VALUES(?, ?, ?, ?)',
            (34, 1, 27, 50),
        )
        self.assert_fsck(self.fsck.check_objects_id)

    def test_objects_hash(self):
        id_ = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                8,
            ),
        )
        self._link(b'test-entry', id_)

        # Assume that due to a crash we did not write the hash for the block
        self.backend['s3ql_data_4364'] = b'Testdata'
        self.db.execute(
            'INSERT INTO objects (id, refcount, phys_size, length) VALUES(?, ?, ?, ?)',
            (4364, 1, 8, 8),
        )
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?, ?, ?)', (id_, 0, 4364)
        )

        # Should pick up wrong hash and delete objects
        self.assert_fsck(
            self.fsck.check_objects_hash,
            self.fsck.check_objects_id,
            self.fsck.check_inode_blocks_obj_id,
        )

        assert not self.db.has_val('SELECT obj_id FROM inode_blocks WHERE inode=?', (id_,))
        inode_p = self.db.get_val('SELECT parent_inode FROM contents_v WHERE inode=?', (id_,))
        lof_id = self.db.get_val(
            "SELECT inode FROM contents_v WHERE name=? AND parent_inode=?",
            (b"lost+found", ROOT_INODE),
        )
        assert inode_p == lof_id

    def test_missing_obj(self):
        obj_id = self.db.rowid(
            'INSERT INTO objects (refcount, phys_size, length) VALUES(1, 32, 128)'
        )

        id_ = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                128,
            ),
        )
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?,?,?)', (id_, 0, obj_id)
        )

        self._link(b'test-entry', id_)
        self.assert_fsck(self.fsck.check_objects_id, self.fsck.check_inode_blocks_obj_id)

    def test_inode_blocks_inode(self):
        obj_id = self.db.rowid(
            'INSERT INTO objects (refcount, phys_size, length, hash) VALUES(1, 42, 34, ?)',
            (sha256(b'foo'),),
        )
        self.backend['s3ql_data_%d' % obj_id] = b'foo'

        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?,?,?)', (27, 0, obj_id)
        )

        self.assert_fsck(self.fsck.check_inode_blocks_inode)

    def test_inode_blocks_obj_id(self):
        id_ = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                128,
            ),
        )
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?,?,?)', (id_, 0, 35)
        )

        self._link(b'test-entry', id_)
        self.assert_fsck(self.fsck.check_inode_blocks_obj_id)

    def test_symlinks_inode(self):
        self.db.execute(
            'INSERT INTO symlink_targets (inode, target) VALUES(?,?)', (42, b'somewhere else')
        )

        self.assert_fsck(self.fsck.check_symlinks_inode)

    def test_ext_attrs_inode(self):
        self.db.execute(
            'INSERT INTO ext_attributes (name_id, inode, value) VALUES(?,?,?)',
            (self._add_name(b'some name'), 34, b'some value'),
        )

        self.assert_fsck(self.fsck.check_ext_attributes_inode)

    def test_ext_attrs_name(self):
        id_ = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                0,
                0,
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                128,
            ),
        )
        self._link(b'test-entry', id_)

        self.db.execute(
            'INSERT INTO ext_attributes (name_id, inode, value) VALUES(?,?,?)',
            (34, id_, b'some value'),
        )

        self.assert_fsck(self.fsck.check_ext_attributes_name)

    @staticmethod
    def random_data(len_):
        with open("/dev/urandom", "rb") as fd:
            return fd.read(len_)

    def test_loops(self):
        # Create some directory inodes
        inodes = [
            self.db.rowid(
                "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
                "VALUES (?,?,?,?,?,?,?)",
                (
                    stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR,
                    0,
                    0,
                    time_ns(),
                    time_ns(),
                    time_ns(),
                    1,
                ),
            )
            for _ in range(3)
        ]

        inodes.append(inodes[0])
        last = inodes[0]
        for inode in inodes[1:]:
            self.db.execute(
                'INSERT INTO contents (name_id, inode, parent_inode) VALUES(?, ?, ?)',
                (self._add_name(str(inode).encode()), inode, last),
            )
            last = inode

        self.assert_fsck(self.fsck.check_loops)

    def test_tmpfile(self):
        # Ensure that path exists
        objname = 's3ql_data_38375'
        self.backend[objname] = b'bla'
        del self.backend[objname]
        path = self.backend._key_to_path(objname)
        tmpname = '%s#%d-%d.tmp' % (path, os.getpid(), _thread.get_ident())
        with open(tmpname, 'wb') as fh:
            fh.write(b'Hello, world')

        self.assert_fsck(self.fsck.check_objects_temp)

    def test_obj_refcounts(self):
        obj_id = self.db.rowid(
            'INSERT INTO objects (refcount, phys_size, length, hash) VALUES(1, 42, 0, ?)',
            (sha256(b'foo'),),
        )
        self.backend['s3ql_data_%d' % obj_id] = b'foo'

        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                2048,
            ),
        )
        self._link(b'test-entry', inode)
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?,?,?)', (inode, 1, obj_id)
        )
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?,?,?)', (inode, 2, obj_id)
        )

        self.assert_fsck(self.fsck.check_objects_refcount)

    def test_orphaned_obj(self):
        self.db.rowid(
            'INSERT INTO objects (refcount, phys_size, length, hash) VALUES(1, 33, 50, ?)',
            (sha256(b'foobar'),),
        )
        self.assert_fsck(self.fsck.check_objects_refcount)

    def test_unix_size(self):
        inode = 42
        self.db.execute(
            "INSERT INTO inodes (id, mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?,?)",
            (
                inode,
                stat.S_IFIFO | stat.S_IRUSR | stat.S_IWUSR,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                0,
            ),
        )
        self._link(b'test-entry', inode)
        self.db.execute('UPDATE inodes SET size = 1 WHERE id=?', (inode,))
        self.assert_fsck(self.fsck.check_unix, can_fix=False)

    def test_unix_size_symlink(self):
        inode = 42
        target = b'some funny random string'
        self.db.execute(
            "INSERT INTO inodes (id, mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
            "VALUES (?,?,?,?,?,?,?,?,?)",
            (
                inode,
                stat.S_IFLNK | stat.S_IRUSR | stat.S_IWUSR,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
                len(target),
            ),
        )
        self.db.execute('INSERT INTO symlink_targets (inode, target) VALUES(?,?)', (inode, target))
        self._link(b'test-entry', inode)
        self.db.execute('UPDATE inodes SET size = 0 WHERE id=?', (inode,))
        self.assert_fsck(self.fsck.check_unix, can_fix=False)

    def test_unix_target(self):
        inode = 42
        self.db.execute(
            "INSERT INTO inodes (id, mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                inode,
                stat.S_IFCHR | stat.S_IRUSR | stat.S_IWUSR,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
            ),
        )
        self._link(b'test-entry', inode)
        self.db.execute('INSERT INTO symlink_targets (inode, target) VALUES(?,?)', (inode, 'foo'))
        self.assert_fsck(self.fsck.check_unix, can_fix=False)

    def test_unix_nomode_reg(self):
        perms = stat.S_IRUSR | stat.S_IWUSR | stat.S_IROTH | stat.S_IRGRP
        stamp = time_ns()
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
            "VALUES (?,?,?,?,?,?,?)",
            (perms, os.getuid(), os.getgid(), stamp, stamp, stamp, 1),
        )
        self._link(b'test-entry', inode)

        self.assert_fsck(self.fsck.check_unix)

        newmode = self.db.get_val('SELECT mode FROM inodes WHERE id=?', (inode,))
        self.assertEqual(stat.S_IMODE(newmode), perms)
        self.assertEqual(stat.S_IFMT(newmode), stat.S_IFREG)

    def test_unix_nomode_dir(self):
        perms = stat.S_IRUSR | stat.S_IWUSR | stat.S_IROTH | stat.S_IRGRP
        stamp = time_ns()
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
            "VALUES (?,?,?,?,?,?,?)",
            (perms, os.getuid(), os.getgid(), stamp, stamp, stamp, 1),
        )
        inode2 = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
            "VALUES (?,?,?,?,?,?,?)",
            (perms | stat.S_IFREG, os.getuid(), os.getgid(), stamp, stamp, stamp, 1),
        )

        self._link(b'test-entry', inode)
        self._link(b'subentry', inode2, inode)

        self.assert_fsck(self.fsck.check_unix)

        newmode = self.db.get_val('SELECT mode FROM inodes WHERE id=?', (inode,))
        self.assertEqual(stat.S_IMODE(newmode), perms)
        self.assertEqual(stat.S_IFMT(newmode), stat.S_IFDIR)

    def test_unix_symlink_no_target(self):
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
            "VALUES (?,?,?,?,?,?,?)",
            (
                stat.S_IFLNK | stat.S_IRUSR | stat.S_IWUSR,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
            ),
        )
        self._link(b'test-entry', inode)
        self.assert_fsck(self.fsck.check_unix, can_fix=False)

    def test_unix_rdev(self):
        inode = 42
        self.db.execute(
            "INSERT INTO inodes (id, mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
            "VALUES (?,?,?,?,?,?,?,?)",
            (
                inode,
                stat.S_IFIFO | stat.S_IRUSR | stat.S_IWUSR,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
            ),
        )
        self._link(b'test-entry', inode)
        self.db.execute('UPDATE inodes SET rdev=? WHERE id=?', (42, inode))
        self.assert_fsck(self.fsck.check_unix, can_fix=False)

    def test_unix_child(self):
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
            "VALUES (?,?,?,?,?,?,?)",
            (
                stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
            ),
        )
        self._link(b'test-entry', inode)
        self.db.execute(
            'INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)',
            (self._add_name(b'foo'), ROOT_INODE, inode),
        )
        self.assert_fsck(self.fsck.check_unix, can_fix=False)

    def test_unix_blocks(self):
        # Socket with data blocks
        inode = self.db.rowid(
            "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
            "VALUES (?,?,?,?,?,?,?)",
            (
                stat.S_IFSOCK | stat.S_IRUSR | stat.S_IWUSR,
                os.getuid(),
                os.getgid(),
                time_ns(),
                time_ns(),
                time_ns(),
                1,
            ),
        )
        self._link(b'test-entry', inode)
        obj_id = self.db.rowid('INSERT INTO objects (refcount, phys_size, length) VALUES(1, 32, 0)')
        self.db.execute(
            'INSERT INTO inode_blocks (inode, blockno, obj_id) VALUES(?,?,?)', (inode, 1, obj_id)
        )

        self.assert_fsck(self.fsck.check_unix, can_fix=False)



#!/usr/bin/env python3
'''
t5_full.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import shutil
import subprocess
import tempfile
from subprocess import CalledProcessError, check_output

import pytest
import t4_fuse
from common import NoTestSection, get_remote_test_info, populate_dir, skip_without_rsync

from s3ql import backends


class TestFull(t4_fuse.TestFuse):
    def populate_dir(self, path):
        populate_dir(path)

    def test(self):
        skip_without_rsync()

        ref_dir = tempfile.mkdtemp(prefix='s3ql-ref-')
        try:
            self.populate_dir(ref_dir)
            self.mkfs()
            self.mount()
            subprocess.check_call(['rsync', '-aHAX', ref_dir + '/', self.mnt_dir + '/'])
            self.umount()
            self.fsck()

            # Delete cache, run fsck and compare
            shutil.rmtree(self.cache_dir)
            self.cache_dir = tempfile.mkdtemp('s3ql-cache-')
            self.fsck()
            self.mount()
            try:
                out = check_output(
                    [
                        'rsync',
                        '-anciHAX',
                        '--delete',
                        '--exclude',
                        '/lost+found',
                        ref_dir + '/',
                        self.mnt_dir + '/',
                    ],
                    universal_newlines=True,
                    stderr=subprocess.STDOUT,
                )
            except CalledProcessError as exc:
                pytest.fail('rsync failed with ' + exc.output)
            if out:
                pytest.fail('Copy not equal to original, rsync says:\n' + out)

            self.umount()

            # Delete cache and mount
            shutil.rmtree(self.cache_dir)
            self.cache_dir = tempfile.mkdtemp(prefix='s3ql-cache-')
            self.mount()
            self.umount()

        finally:
            shutil.rmtree(ref_dir)


class RemoteTest:
    def setup_method(self, method, name):
        super().setup_method(method)
        try:
            (backend_login, backend_pw, self.storage_url) = get_remote_test_info(name)
        except NoTestSection as exc:
            super().teardown_method(method)
            pytest.skip(exc.reason)
        self.backend_login = backend_login
        self.backend_passphrase = backend_pw

    def populate_dir(self, path):
        populate_dir(path, entries=50, size=5 * 1024 * 1024)

    def teardown_method(self, method):
        super().teardown_method(method)

        proc = subprocess.Popen(
            self.s3ql_cmd_argv('s3qladm')
            + ['--quiet', '--authfile', '/dev/null', 'clear', self.storage_url],
            stdin=subprocess.PIPE,
            universal_newlines=True,
        )
        if self.backend_login is not None:
            print(self.backend_login, file=proc.stdin)
            print(self.backend_passphrase, file=proc.stdin)
        print('yes', file=proc.stdin)
        proc.stdin.close()

        assert proc.wait() == 0


# Dynamically generate tests for other backends
for backend_name in backends.prefix_map:
    if backend_name == 'local':
        continue

    def setup_method(self, method, backend_name=backend_name):
        RemoteTest.setup_method(self, method, backend_name + '-test')

    test_class_name = 'TestFull' + backend_name
    globals()[test_class_name] = type(
        test_class_name, (RemoteTest, TestFull), {'setup_method': setup_method}
    )



#!/usr/bin/env python3
'''
t5_cache.py - this file is part of S3QL.

Copyright © 2017 IOFabric, Inc.

All Rights Reserved.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import os
import shutil
import subprocess
import tempfile
import time
from os.path import join as pjoin

import pytest
import t4_fuse

from s3ql.common import escape

with open(__file__, 'rb') as fh:
    TEST_DATA = fh.read()


class TestPerstCache(t4_fuse.TestFuse):
    # Need to overwrite parent class' method with something, thus the
    # inexpressive name.
    def test(self):
        # Write test data
        self.mkfs()
        self.mount(extra_args=['--keep-cache'])
        with open(pjoin(self.mnt_dir, 'testfile'), 'wb') as fh:
            fh.write(TEST_DATA)
        self.umount()

        # Poison backend storage object
        with open(pjoin(self.backend_dir, 's3ql_data_', 's3ql_data_1'), 'wb') as fh:
            fh.write(b'poison!')

        # fsck.s3ql shouldn't report errors
        self.fsck()

        # Ensure that we read from cache
        self.mount()
        with open(pjoin(self.mnt_dir, 'testfile'), 'rb') as fh:
            assert fh.read() == TEST_DATA
        self.umount()

    def test_cache_upload(self):
        self.reg_output(r'^WARNING: Writing dirty block', count=1)
        self.reg_output(r'^WARNING: Remote metadata is outdated', count=1)

        # Write test data
        self.mkfs()
        self.mount(extra_args=['--keep-cache'])
        with open(pjoin(self.mnt_dir, 'testfile'), 'wb') as fh:
            fh.write(TEST_DATA)

        # Kill mount
        self.flush_cache()
        self.mount_process.kill()
        self.mount_process.wait()
        self.umount_fuse()

        # Update cache files
        cache_file = pjoin(self.cache_dir, escape(self.storage_url) + '-cache', '4-0')
        with open(cache_file, 'rb+') as fh:
            fh.write(TEST_DATA[::-1])
        self.fsck(expect_retcode=128)
        assert not os.path.exists(cache_file)

        # Ensure that we get updated data
        self.reg_output(r'^WARNING: sqlite3: recovered [0-9]+ frames from WAL file', count=1)
        self.mount()
        with open(pjoin(self.mnt_dir, 'testfile'), 'rb') as fh:
            assert fh.read() == TEST_DATA[::-1]
        self.umount()

    def upload_meta(self):
        subprocess.check_call(
            self.s3ql_cmd_argv('s3qlctrl') + ['--quiet', 'backup-metadata', self.mnt_dir]
        )

    # Check that cache is ignored if fs was mounted elsewhere
    @pytest.mark.parametrize("with_fsck", (True, False))
    def test_cache_flush(self, with_fsck):
        # Write test data
        self.mkfs()
        self.mount(extra_args=['--keep-cache'])
        with open(pjoin(self.mnt_dir, 'testfile'), 'wb') as fh:
            fh.write(TEST_DATA)
            ino = os.fstat(fh.fileno()).st_ino
        self.umount()

        # Taint cache file, so we'll get an error if it's re-used
        cache_file = pjoin(self.cache_dir, escape(self.storage_url) + '-cache', '%d-0' % (ino,))
        with open(cache_file, 'rb+') as fh:
            fh.write(TEST_DATA[::-1])

        # Mount elsewhere
        bak = self.cache_dir
        self.cache_dir = tempfile.mkdtemp(prefix='s3ql-cache-')
        try:
            self.mount()
            with open(pjoin(self.mnt_dir, 'testfile2'), 'wb') as fh:
                fh.write(b'hello')
            self.umount()
        finally:
            shutil.rmtree(self.cache_dir)
            self.cache_dir = bak

        # Make sure that cache is ignored
        if with_fsck:
            self.fsck(args=['--keep-cache'])
        self.mount()
        with open(pjoin(self.mnt_dir, 'testfile'), 'rb') as fh:
            assert fh.read() == TEST_DATA
        self.umount()

    # Check that cache is ignored if fs was mounted elsewhere
    # and was not cleanly unmounted on either system
    def test_cache_flush_unclean(self):
        self.reg_output(r'^WARNING: Renaming outdated cache directory', count=1)
        self.reg_output(r'^WARNING: You should delete this directory', count=1)

        # Write test data
        self.mkfs()
        self.mount(extra_args=['--keep-cache'])
        with open(pjoin(self.mnt_dir, 'testfile'), 'wb') as fh:
            fh.write(TEST_DATA)

        # Kill mount
        self.flush_cache()
        self.upload_meta()

        # Metadata upload is not blocking, so give it some time to
        # complete
        time.sleep(1)

        self.mount_process.kill()
        self.mount_process.wait()
        self.umount_fuse()

        # Update cache files
        cache_file = pjoin(self.cache_dir, escape(self.storage_url) + '-cache', '4-0')
        with open(cache_file, 'rb+') as fh:
            fh.write(TEST_DATA[::-1])

        # Mount elsewhere
        self.reg_output(r'^WARNING: sqlite3: recovered [0-9]+ frames from WAL file', count=1)
        bak = self.cache_dir
        self.cache_dir = tempfile.mkdtemp(prefix='s3ql-cache-')
        try:
            self.fsck(expect_retcode=0, args=['--force-remote'])
            self.mount()
            with open(pjoin(self.mnt_dir, 'testfile2'), 'wb') as fh:
                fh.write(b'hello')
            self.mount_process.kill()
            self.mount_process.wait()
            self.umount_fuse()
        finally:
            shutil.rmtree(self.cache_dir)
            self.cache_dir = bak

        # Make sure that cache is ignored
        self.fsck(expect_retcode=0, args=['--force-remote', '--keep-cache'])
        self.mount()
        with open(pjoin(self.mnt_dir, 'testfile'), 'rb') as fh:
            assert fh.read() == TEST_DATA
        self.umount()



'''
mkfs.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import atexit
import logging
import os
import shutil
import stat
import sys
import time
from base64 import b64encode
from getpass import getpass

from . import CTRL_INODE, CURRENT_FS_REV, ROOT_INODE
from .backends import s3
from .backends.comprenc import ComprencBackend
from .common import get_backend, split_by_n, time_ns
from .database import (
    Connection,
    FsAttributes,
    create_tables,
    upload_metadata,
    upload_params,
    write_params,
)
from .logging import QuietError, setup_logging, setup_warnings
from .parse_args import ArgumentParser

log = logging.getLogger(__name__)


def parse_args(args):
    parser = ArgumentParser(description="Initializes an S3QL file system")

    parser.add_cachedir()
    parser.add_log()
    parser.add_debug()
    parser.add_quiet()
    parser.add_backend_options()
    parser.add_version()
    parser.add_storage_url()

    parser.add_argument(
        "-L",
        default='',
        help="Filesystem label",
        dest="label",
        metavar='<name>',
    )
    parser.add_argument(
        "--data-block-size",
        type=int,
        default=1024,
        metavar='<size>',
        help="Block size (in KiB) to use for storing file contents. Files "
        "larger than this size will be stored in multiple backend objects. "
        "Backend object size may be smaller than this due to compression. "
        "Default: %(default)d KiB.",
    )
    parser.add_argument(
        "--metadata-block-size",
        type=int,
        default=64,
        metavar='<size>',
        help="Block size (in KiB) to use for storing filesystem metadata. "
        "Backend object size may be smaller than this due to compression. "
        "Default: %(default)d KiB.",
    )
    parser.add_argument(
        "--plain", action="store_true", default=False, help="Create unencrypted file system."
    )

    options = parser.parse_args(args)

    return options


def init_tables(conn):
    # Insert root directory
    now_ns = time_ns()
    conn.execute(
        "INSERT INTO inodes (id,mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
        "VALUES (?,?,?,?,?,?,?,?)",
        (
            ROOT_INODE,
            stat.S_IFDIR
            | stat.S_IRUSR
            | stat.S_IWUSR
            | stat.S_IXUSR
            | stat.S_IRGRP
            | stat.S_IXGRP
            | stat.S_IROTH
            | stat.S_IXOTH,
            os.getuid(),
            os.getgid(),
            now_ns,
            now_ns,
            now_ns,
            1,
        ),
    )

    # Insert control inode, the actual values don't matter that much
    conn.execute(
        "INSERT INTO inodes (id,mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
        "VALUES (?,?,?,?,?,?,?,?)",
        (CTRL_INODE, stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR, 0, 0, now_ns, now_ns, now_ns, 42),
    )

    # Insert lost+found directory
    inode = conn.rowid(
        "INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount) "
        "VALUES (?,?,?,?,?,?,?)",
        (
            stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR,
            os.getuid(),
            os.getgid(),
            now_ns,
            now_ns,
            now_ns,
            1,
        ),
    )
    name_id = conn.rowid('INSERT INTO names (name, refcount) VALUES(?,?)', (b'lost+found', 1))
    conn.execute(
        "INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)",
        (name_id, inode, ROOT_INODE),
    )


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    options = parse_args(args)
    setup_logging(options)

    setup_warnings()
    plain_backend = get_backend(options, raw=True)
    atexit.register(plain_backend.close)

    log.info(
        "Before using S3QL, make sure to read the user's guide, especially\n"
        "the 'Important Rules to Avoid Losing Data' section."
    )

    if isinstance(plain_backend, s3.Backend) and '.' in plain_backend.bucket_name:
        log.warning(
            'S3 Buckets with names containing dots cannot be '
            'accessed using SSL!'
            '(cf. https://forums.aws.amazon.com/thread.jspa?threadID=130560)'
        )

    if 's3ql_params' in plain_backend or 's3ql_metadata' in plain_backend:
        raise QuietError(
            "Refusing to overwrite existing file system! (use `s3qladm clear` to delete)"
        )

    if not options.plain:
        if sys.stdin.isatty():
            wrap_pw = getpass("Enter encryption password: ")
            if not wrap_pw == getpass("Confirm encryption password: "):
                raise QuietError("Passwords don't match.")
        else:
            wrap_pw = sys.stdin.readline().rstrip()
        wrap_pw = wrap_pw.encode('utf-8')

        # Generate data encryption passphrase
        log.info('Generating random encryption key...')
        fh = open('/dev/urandom', "rb", 0)  # No buffering
        data_pw = fh.read(32)
        fh.close()

        backend = ComprencBackend(wrap_pw, ('lzma', 2), plain_backend)
        backend['s3ql_passphrase'] = data_pw
        backend['s3ql_passphrase_bak1'] = data_pw
        backend['s3ql_passphrase_bak2'] = data_pw
        backend['s3ql_passphrase_bak3'] = data_pw
    else:
        data_pw = None

    backend = ComprencBackend(data_pw, ('lzma', 2), plain_backend)
    atexit.unregister(plain_backend.close)
    atexit.register(backend.close)
    cachepath = options.cachepath

    # There can't be a corresponding backend, so we can safely delete
    # these files.
    if os.path.exists(cachepath + '.db'):
        os.unlink(cachepath + '.db')
    if os.path.exists(cachepath + '-cache'):
        shutil.rmtree(cachepath + '-cache')

    param = FsAttributes(
        revision=CURRENT_FS_REV,
        seq_no=int(time.time()),
        label=options.label,
        data_block_size=options.data_block_size * 1024,
        metadata_block_size=options.metadata_block_size * 1024,
        last_fsck=time.time(),
        last_modified=time.time(),
    )

    log.info('Creating metadata tables...')
    db = Connection(cachepath + '.db', param.metadata_block_size)
    create_tables(db)
    init_tables(db)

    log.info('Uploading metadata...')
    db.close()
    upload_metadata(backend, db, param)
    write_params(cachepath, param)
    upload_params(backend, param)
    if os.path.exists(cachepath + '-cache'):
        shutil.rmtree(cachepath + '-cache')

    if data_pw is not None:
        print(
            'Please store the following master key in a safe location. It allows ',
            'decryption of the S3QL file system in case the storage objects holding ',
            'this information get corrupted:',
            '---BEGIN MASTER KEY---',
            ' '.join(split_by_n(b64encode(data_pw).decode(), 4)),
            '---END MASTER KEY---',
            sep='\n',
        )


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3
'''
t5_cp.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import os.path
import shutil
import subprocess
import sys
import tempfile
from subprocess import CalledProcessError, check_output

import pytest
import t4_fuse
from common import populate_dir, retry, skip_without_rsync


class TestCp(t4_fuse.TestFuse):
    def test(self):
        skip_without_rsync()
        self.mkfs()
        self.mount()
        self.tst_cp()
        self.umount()
        self.fsck()

    def test_bg(self):
        skip_without_rsync()
        self.mkfs()
        self.logfile = tempfile.NamedTemporaryFile()
        self.mount(in_foreground=False, extra_args=["--log", self.logfile.name])
        self.tst_cp()
        self.umount()
        self.fsck()
        self.logfile.close()

    def tst_cp(self):
        tempdir = tempfile.mkdtemp(prefix='s3ql-cp-')
        try:
            populate_dir(tempdir)

            # Rsync
            subprocess.check_call(
                ['rsync', '-aHAX', tempdir + '/', os.path.join(self.mnt_dir, 'orig') + '/']
            )

            # copy
            subprocess.check_call(
                self.s3ql_cmd_argv('s3qlcp')
                + [
                    '--quiet',
                    os.path.join(self.mnt_dir, 'orig'),
                    os.path.join(self.mnt_dir, 'copy'),
                ]
            )

            # compare
            try:
                out = check_output(
                    [
                        'rsync',
                        '-anciHAX',
                        '--delete',
                        tempdir + '/',
                        os.path.join(self.mnt_dir, 'copy') + '/',
                    ],
                    universal_newlines=True,
                    stderr=subprocess.STDOUT,
                )
            except CalledProcessError as exc:
                pytest.fail('rsync failed with ' + exc.output)

            if out:
                pytest.fail('Copy not equal to original, rsync says:\n' + out)

        finally:
            shutil.rmtree(tempdir)

    def test_cp_inode_invalidate(self):
        # check if we can write to drop_caches
        try:
            open("/proc/sys/vm/drop_caches", "w").close()
        except OSError:
            pytest.skip('test_cp_inode_invalidate requires drop_caches to be writable, skipping.')

        self.passphrase = None
        self.mkfs()

        # Run monkeypatched mount.s3ql with overridden pyfuse3.invalidate_inode :
        # Drop kernel dentries and inodes cache just before calling pyfuse3.invalidate_inode
        cmd = [
            sys.executable,
            os.path.join(os.path.dirname(__file__), 'mount_helper.py'),
            "--fg",
            '--cachedir',
            self.cache_dir,
            '--log',
            'none',
            '--compress',
            'zlib',
            '--quiet',
            self.storage_url,
            self.mnt_dir,
            '--authfile',
            '/dev/null',
        ]
        self.mount_process = subprocess.Popen(cmd, universal_newlines=True)

        def poll():
            if os.path.ismount(self.mnt_dir):
                return True
            assert self.mount_process.poll() is None

        retry(10, poll)

        os.mkdir(os.path.join(self.mnt_dir, 'orig'))

        cmd = self.s3ql_cmd_argv('s3qlcp') + [
            '--quiet',
            os.path.join(self.mnt_dir, 'orig'),
            os.path.join(self.mnt_dir, 'copy'),
        ]
        cp_process = subprocess.Popen(cmd)
        retry(5, lambda: cp_process.poll() is not None)
        assert cp_process.wait() == 0

        self.umount()
        self.fsck()



'''
cp.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import logging
import os
import stat
import sys
import textwrap

import pyfuse3

from .common import assert_fs_owner
from .logging import QuietError, setup_logging, setup_warnings
from .parse_args import ArgumentParser

log = logging.getLogger(__name__)


def parse_args(args):
    '''Parse command line'''

    parser = ArgumentParser(
        description=textwrap.dedent(
            '''\
        Replicates the contents of the directory <source> in the
        directory <target>. <source> has to be an existing directory and
        <target>  must not exist. Both directories have to be within
        the same S3QL file system.

        The replication will not take any additional space. Only if one
        of directories is modified later on, the modified data will take
        additional storage space.
        '''
        )
    )

    parser.add_log()
    parser.add_debug()
    parser.add_quiet()
    parser.add_version()

    parser.add_argument('source', help='source directory', type=(lambda x: x.rstrip('/')))
    parser.add_argument('target', help='target directory', type=(lambda x: x.rstrip('/')))

    options = parser.parse_args(args)

    return options


def main(args=None):
    '''Efficiently copy a directory tree'''

    if args is None:
        args = sys.argv[1:]

    setup_warnings()
    options = parse_args(args)
    setup_logging(options)

    if not os.path.exists(options.source):
        raise QuietError('Source directory %r does not exist' % options.source)

    if os.path.exists(options.target):
        raise QuietError('Target directory %r must not yet exist.' % options.target)

    parent = os.path.dirname(os.path.abspath(options.target))
    if not os.path.exists(parent):
        raise QuietError('Target parent %r does not exist' % parent)

    fstat_s = os.stat(options.source)
    fstat_p = os.stat(parent)
    if not stat.S_ISDIR(fstat_s.st_mode):
        raise QuietError('Source %r is not a directory' % options.source)

    if not stat.S_ISDIR(fstat_p.st_mode):
        raise QuietError('Target parent %r is not a directory' % parent)

    if fstat_p.st_dev != fstat_s.st_dev:
        raise QuietError('Source and target are not on the same file system.')

    if os.path.ismount(options.source):
        raise QuietError('%s is a mount point.' % options.source)

    ctrlfile = assert_fs_owner(options.source)
    try:
        os.mkdir(options.target)
    except PermissionError:
        raise QuietError('No permission to create target directory')

    # Make sure that write cache is flushed
    pyfuse3.syncfs(options.target)

    # Ensure the inode of the target folder stays in the kernel dentry cache
    # (We invalidate it during the copy)
    with os.scandir(options.target):
        fstat_t = os.stat(options.target)
        pyfuse3.setxattr(ctrlfile, 'copy', ('(%d, %d)' % (fstat_s.st_ino, fstat_t.st_ino)).encode())


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3
'''
pcp.py - this file is part of S3QL.

Parallel, recursive copy of directory trees.

---
Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import logging
import os
import subprocess
import sys
from argparse import ArgumentTypeError

# We are running from the S3QL source directory, make sure
# that we use modules from this directory
basedir = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), '..'))
if os.path.exists(os.path.join(basedir, 'setup.py')) and os.path.exists(
    os.path.join(basedir, 'src', 's3ql', '__init__.py')
):
    sys.path = [os.path.join(basedir, 'src')] + sys.path

from s3ql.logging import setup_logging, setup_warnings
from s3ql.parse_args import ArgumentParser

log = logging.getLogger(__name__)

pool = ('abcdefghijklmnopqrstuvwxyz', 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', '0123456789')


def processes_to_int(string):
    value = int(string)
    if value < 2 or value > len(pool[0]) + 1:
        raise ArgumentTypeError(
            "%d needs to be between 2 and %d (inclusive)" % (value, len(pool[0]) + 1)
        )
    return value


def parse_args(args):
    '''Parse command line'''

    parser = ArgumentParser(
        description='Recursively copy source(s) to destination using multiple '
        'parallel rsync processes.'
    )

    parser.add_log()
    parser.add_quiet()
    parser.add_debug()
    parser.add_version()

    parser.add_argument("-a", action="store_true", help='Pass -aHAX option to rsync.')
    parser.add_argument(
        "--processes",
        action="store",
        type=processes_to_int,
        metavar='<no>',
        default=10,
        help='Number of rsync processes to use (default: %(default)s).',
    )

    parser.add_argument('source', metavar='<source>', nargs='+', help='Directories to copy')
    parser.add_argument('dest', metavar='<destination>', help="Target directory")

    options = parser.parse_args(args)
    options.pps = options.source + [options.dest]

    return options


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    setup_warnings()
    options = parse_args(args)
    setup_logging(options)

    steps = [len(x) / (options.processes - 1) for x in pool]
    prefixes = list()
    for i in range(options.processes - 1):
        parts = [x[int(i * y) : int((i + 1) * y)] for (x, y) in zip(pool, steps)]
        prefixes.append(''.join(parts))

    filters = ['-! [%s]*' % x for x in prefixes]

    # Catch all
    filters.append('- [%s]*' % ''.join(prefixes))

    rsync_args = ['rsync', '-f', '+ */']
    if not options.quiet:
        rsync_args.append('--out-format')
        rsync_args.append('%n%L')
    if options.a:
        rsync_args.append('-aHAX')

    processes = list()
    for filter_ in filters:
        cmd = rsync_args + ['-f', filter_] + options.pps
        log.debug('Calling %s', cmd)
        processes.append(subprocess.Popen(cmd))

    if all([c.wait() == 0 for c in processes]):
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3
'''
mock_server.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import hashlib
import json
import logging
import re
import socketserver
import urllib.parse
from http.server import BaseHTTPRequestHandler
from xml.sax.saxutils import escape as xml_escape

log = logging.getLogger(__name__)

ERROR_RESPONSE_TEMPLATE = '''\
<?xml version="1.0" encoding="UTF-8"?>
<Error>
  <Code>%(code)s</Code>
  <Message>%(message)s</Message>
  <Resource>%(resource)s</Resource>
  <RequestId>%(request_id)s</RequestId>
</Error>
'''

COPY_RESPONSE_TEMPLATE = '''\
<?xml version="1.0" encoding="UTF-8"?>
<CopyObjectResult xmlns="%(ns)s">
   <LastModified>2008-02-20T22:13:01</LastModified>
   <ETag>&quot;%(etag)s&quot;</ETag>
</CopyObjectResult>
'''


class StorageServer(socketserver.TCPServer):
    def __init__(self, request_handler, server_address):
        super().__init__(server_address, request_handler)
        self.data = dict()
        self.metadata = dict()
        self.hostname = self.server_address[0]
        self.port = self.server_address[1]


class ParsedURL:
    __slots__ = ['bucket', 'key', 'params', 'fragment']


class S3CRequestHandler(BaseHTTPRequestHandler):
    '''A request handler implementing a subset of the AWS S3 Interface

    Bucket names are ignored, all keys share the same global
    namespace.
    '''

    server_version = "MockHTTP"
    protocol_version = 'HTTP/1.1'
    meta_header_re = re.compile(r'X-AMZ-Meta-([a-z0-9_.-]+)$', re.IGNORECASE)
    hdr_prefix = 'X-AMZ-'
    xml_ns = 'http://s3.amazonaws.com/doc/2006-03-01/'

    def parse_url(self, path):
        p = ParsedURL()
        q = urllib.parse.urlsplit(path)

        path = urllib.parse.unquote(q.path)

        assert path[0] == '/'
        (p.bucket, p.key) = path[1:].split('/', maxsplit=1)

        p.params = urllib.parse.parse_qs(q.query)
        p.fragment = q.fragment

        return p

    def log_message(self, format, *args):
        log.debug(format, *args)

    def handle(self):
        # Ignore exceptions resulting from the client closing
        # the connection.
        try:
            return super().handle()
        except ValueError as exc:
            if exc.args == ('I/O operation on closed file.',):
                pass
            else:
                raise
        except (BrokenPipeError, ConnectionResetError):
            pass

    def do_DELETE(self):
        q = self.parse_url(self.path)
        try:
            del self.server.data[q.key]
            del self.server.metadata[q.key]
        except KeyError:
            self.send_error(404, code='NoSuchKey', resource=q.key)
            return
        else:
            self.send_response(204)
            self.end_headers()

    def _check_encoding(self):
        encoding = self.headers['Content-Encoding']
        if 'Content-Length' not in self.headers:
            self.send_error(400, message='Missing Content-Length', code='MissingContentLength')
            return
        elif encoding and encoding != 'identity':
            self.send_error(501, message='Unsupported encoding', code='NotImplemented')
            return

        return int(self.headers['Content-Length'])

    def _get_meta(self):
        meta = dict()
        for name, value in self.headers.items():
            hit = self.meta_header_re.search(name)
            if hit:
                meta[hit.group(1)] = value
        return meta

    def do_PUT(self):
        len_ = self._check_encoding()
        q = self.parse_url(self.path)
        meta = self._get_meta()

        src = self.headers.get(self.hdr_prefix + 'copy-source')
        if src and len_:
            self.send_error(
                400, message='Upload and copy are mutually exclusive', code='UnexpectedContent'
            )
            return
        elif src:
            src = urllib.parse.unquote(src)
            hit = re.match('^/([a-z0-9._-]+)/(.+)$', src)
            if not hit:
                self.send_error(400, message='Cannot parse copy-source', code='InvalidArgument')
                return

            metadata_directive = self.headers.get(self.hdr_prefix + 'metadata-directive', 'COPY')
            if metadata_directive not in ('COPY', 'REPLACE'):
                self.send_error(400, message='Invalid metadata directive', code='InvalidArgument')
                return
            src = hit.group(2)
            try:
                data = self.server.data[src]
                self.server.data[q.key] = data
                if metadata_directive == 'COPY':
                    self.server.metadata[q.key] = self.server.metadata[src]
                else:
                    self.server.metadata[q.key] = meta
            except KeyError:
                self.send_error(404, code='NoSuchKey', resource=src)
                return
        else:
            data = self.rfile.read(len_)
            self.server.metadata[q.key] = meta
            self.server.data[q.key] = data

        md5 = hashlib.md5()
        md5.update(data)

        if src:
            content = (
                COPY_RESPONSE_TEMPLATE % {'etag': md5.hexdigest(), 'ns': self.xml_ns}
            ).encode('utf-8')
            self.send_response(200)
            self.send_header('ETag', '"%s"' % md5.hexdigest())
            self.send_header('Content-Length', str(len(content)))
            self.send_header("Content-Type", 'text/xml')
            self.end_headers()
            self.wfile.write(content)
        else:
            self.send_response(201)
            self.send_header('ETag', '"%s"' % md5.hexdigest())
            self.send_header('Content-Length', '0')
            self.end_headers()

    def handle_expect_100(self):
        if self.command == 'PUT':
            self._check_encoding()

        self.send_response_only(100)
        self.end_headers()
        return True

    def do_GET(self):
        q = self.parse_url(self.path)
        if not q.key:
            return self.do_list(q)

        try:
            data = self.server.data[q.key]
            meta = self.server.metadata[q.key]
        except KeyError:
            self.send_error(404, code='NoSuchKey', resource=q.key)
            return

        self.send_response(200)
        self.send_header("Content-Type", 'application/octet-stream')
        self.send_header("Content-Length", str(len(data)))
        for name, value in meta.items():
            self.send_header(self.hdr_prefix + 'Meta-%s' % name, value)
        md5 = hashlib.md5()
        md5.update(data)
        self.send_header('ETag', '"%s"' % md5.hexdigest())
        self.end_headers()
        self.send_data(data)

    def send_data(self, data):
        self.wfile.write(data)

    def do_list(self, q):
        marker = q.params['marker'][0] if 'marker' in q.params else None
        max_keys = int(q.params['max_keys'][0]) if 'max_keys' in q.params else 1000
        prefix = q.params['prefix'][0] if 'prefix' in q.params else ''

        resp = [
            '<?xml version="1.0" encoding="UTF-8"?>',
            '<ListBucketResult xmlns="%s">' % self.xml_ns,
            '<MaxKeys>%d</MaxKeys>' % max_keys,
            '<IsTruncated>false</IsTruncated>',
        ]

        count = 0
        for key in sorted(self.server.data):
            if not key.startswith(prefix):
                continue
            if marker and key <= marker:
                continue
            resp.append('<Contents><Key>%s</Key></Contents>' % xml_escape(key))
            count += 1
            if count == max_keys:
                resp[3] = '<IsTruncated>true</IsTruncated>'
                break

        resp.append('</ListBucketResult>')
        body = '\n'.join(resp).encode()

        self.send_response(200)
        self.send_header("Content-Type", 'text/xml')
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)

    def do_HEAD(self):
        q = self.parse_url(self.path)
        try:
            meta = self.server.metadata[q.key]
            data = self.server.data[q.key]
        except KeyError:
            self.send_error(404, code='NoSuchKey', resource=q.key)
            return

        self.send_response(200)
        self.send_header("Content-Type", 'application/octet-stream')
        self.send_header("Content-Length", str(len(data)))
        for name, value in meta.items():
            self.send_header(self.hdr_prefix + 'Meta-%s' % name, value)
        self.end_headers()

    def send_error(self, status, message=None, code='', resource='', extra_headers=None):
        if not message:
            try:
                (_, message) = self.responses[status]
            except KeyError:
                message = 'Unknown'

        self.log_error("code %d, message %s", status, message)
        content = (
            ERROR_RESPONSE_TEMPLATE
            % {
                'code': code,
                'message': xml_escape(message),
                'request_id': 42,
                'resource': xml_escape(resource),
            }
        ).encode('utf-8', 'replace')
        self.send_response(status, message)
        self.send_header("Content-Type", 'text/xml; charset="utf-8"')
        self.send_header("Content-Length", str(len(content)))
        if extra_headers:
            for name, value in extra_headers.items():
                self.send_header(name, value)
        self.end_headers()
        if self.command != 'HEAD' and status >= 200 and status not in (204, 304):
            self.wfile.write(content)


class S3C4RequestHandler(S3CRequestHandler):
    '''Request Handler for s3c4 backend

    Currently identical to S3CRequestHandler since mock request handlers
    do not check request signatures.
    '''

    pass


class BasicSwiftRequestHandler(S3CRequestHandler):
    '''A request handler implementing a subset of the OpenStack Swift Interface

    Container and AUTH_* prefix are ignored, all keys share the same global
    namespace.

    To keep it simple, this handler is both storage server and authentication
    server in one.
    '''

    meta_header_re = re.compile(r'X-Object-Meta-([a-z0-9_.-]+)$', re.IGNORECASE)
    hdr_prefix = 'X-Object-'

    SWIFT_INFO = {
        "swift": {
            "max_meta_count": 90,
            "max_meta_value_length": 256,
            "container_listing_limit": 10000,
            "extra_header_count": 0,
            "max_meta_overall_size": 4096,
            "version": "2.0.0",  # < 2.8
            "max_meta_name_length": 128,
            "max_header_size": 16384,
        }
    }

    def parse_url(self, path):
        p = ParsedURL()
        q = urllib.parse.urlsplit(path)

        path = urllib.parse.unquote(q.path)

        assert path[0:4] == '/v1/'
        (_, p.bucket, p.key) = path[4:].split('/', maxsplit=2)

        p.params = urllib.parse.parse_qs(q.query, True)
        p.fragment = q.fragment

        return p

    def do_PUT(self):
        len_ = self._check_encoding()
        q = self.parse_url(self.path)
        meta = self._get_meta()

        src = self.headers.get('x-copy-from')
        if src and len_:
            self.send_error(
                400, message='Upload and copy are mutually exclusive', code='UnexpectedContent'
            )
            return
        elif src:
            src = urllib.parse.unquote(src)
            hit = re.match('^/([a-z0-9._-]+)/(.+)$', src)
            if not hit:
                self.send_error(400, message='Cannot parse x-copy-from', code='InvalidArgument')
                return

            src = hit.group(2)
            try:
                data = self.server.data[src]
                self.server.data[q.key] = data
                if 'x-fresh-metadata' in self.headers:
                    self.server.metadata[q.key] = meta
                else:
                    self.server.metadata[q.key] = self.server.metadata[src].copy()
                    self.server.metadata[q.key].update(meta)
            except KeyError:
                self.send_error(404, code='NoSuchKey', resource=src)
                return
        else:
            data = self.rfile.read(len_)
            self.server.metadata[q.key] = meta
            self.server.data[q.key] = data

        md5 = hashlib.md5()
        md5.update(data)

        if src:
            self.send_response(202)
            self.send_header('X-Copied-From', self.headers['x-copy-from'])
            self.send_header('Content-Length', '0')
            self.end_headers()
        else:
            self.send_response(201)
            self.send_header('ETag', '"%s"' % md5.hexdigest())
            self.send_header('Content-Length', '0')
            self.end_headers()

    def do_POST(self):
        q = self.parse_url(self.path)
        meta = self._get_meta()

        if q.key not in self.server.metadata:
            self.send_error(404, code='NoSuchKey', resource=q.key)
            return

        self.server.metadata[q.key] = meta

        self.send_response(204)
        self.send_header('Content-Length', '0')
        self.end_headers()

    def do_GET(self):
        if self.path in ('/v1.0', '/auth/v1.0'):
            self.send_response(200)
            self.send_header(
                'X-Storage-Url',
                'http://%s:%d/v1/AUTH_xyz' % (self.server.hostname, self.server.port),
            )
            self.send_header('X-Auth-Token', 'static')
            self.send_header('Content-Length', '0')
            self.end_headers()
        elif self.path == '/info':
            content = json.dumps(self.SWIFT_INFO).encode('utf-8')
            self.send_response(200)
            self.send_header('Content-Length', str(len(content)))
            self.send_header("Content-Type", 'application/json; charset="utf-8"')
            self.end_headers()
            self.wfile.write(content)
        else:
            return super().do_GET()

    def do_list(self, q):
        marker = q.params['marker'][0] if 'marker' in q.params else None
        max_keys = int(q.params['limit'][0]) if 'limit' in q.params else 10000
        prefix = q.params['prefix'][0] if 'prefix' in q.params else ''

        resp = []

        count = 0
        for key in sorted(self.server.data):
            if not key.startswith(prefix):
                continue
            if marker and key <= marker:
                continue
            resp.append({'name': key})
            count += 1
            if count == max_keys:
                break

        body = json.dumps(resp).encode('utf-8')

        self.send_response(200)
        self.send_header("Content-Type", 'application/json; charset="utf-8"')
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)


class CopySwiftRequestHandler(BasicSwiftRequestHandler):
    '''OpenStack Swift handler that emulates Copy middleware.'''

    SWIFT_INFO = {
        "swift": {
            "max_meta_count": 90,
            "max_meta_value_length": 256,
            "container_listing_limit": 10000,
            "extra_header_count": 0,
            "max_meta_overall_size": 4096,
            "version": "2.9.0",  # >= 2.8
            "max_meta_name_length": 128,
            "max_header_size": 16384,
        }
    }

    def do_COPY(self):
        src = self.parse_url(self.path)
        meta = self._get_meta()

        try:
            dst = self.headers['destination']
            assert dst[0] == '/'
            (_, dst) = dst[1:].split('/', maxsplit=1)
        except KeyError:
            self.send_error(400, message='No Destination provided', code='InvalidArgument')
            return

        if src.key not in self.server.metadata:
            self.send_error(404, code='NoSuchKey', resource=src)
            return

        if 'x-fresh-metadata' in self.headers:
            self.server.metadata[dst] = meta
        else:
            self.server.metadata[dst] = self.server.metadata[src.key].copy()
            self.server.metadata[dst].update(meta)

        self.server.data[dst] = self.server.data[src.key]

        self.send_response(202)
        self.send_header('X-Copied-From', '%s/%s' % (src.bucket, src.key))
        self.send_header('Content-Length', '0')
        self.end_headers()


class BulkDeleteSwiftRequestHandler(BasicSwiftRequestHandler):
    '''OpenStack Swift handler that emulates bulk middleware (the delete part).'''

    MAX_DELETES = 8  # test deletes 16 objects, so needs two requests
    SWIFT_INFO = {
        "bulk_delete": {"max_failed_deletes": MAX_DELETES, "max_deletes_per_request": MAX_DELETES},
        "swift": {
            "max_meta_count": 90,
            "max_meta_value_length": 256,
            "container_listing_limit": 10000,
            "extra_header_count": 0,
            "max_meta_overall_size": 4096,
            "version": "2.0.0",  # < 2.8
            "max_meta_name_length": 128,
            "max_header_size": 16384,
        },
    }

    def do_POST(self):
        q = self.parse_url(self.path)
        if not 'bulk-delete' in q.params:
            return super().do_POST()

        response = {
            'Response Status': '200 OK',
            'Response Body': '',
            'Number Deleted': 0,
            'Number Not Found': 0,
            'Errors': [],
        }

        def send_response(status_int):
            content = json.dumps(response).encode('utf-8')
            self.send_response(status_int)
            self.send_header('Content-Length', str(len(content)))
            self.send_header("Content-Type", 'application/json; charset="utf-8"')
            self.end_headers()
            self.wfile.write(content)

        def error(reason):
            response['Response Status'] = '502 Internal Server Error'
            response['Response Body'] = reason
            send_response(502)

        def inline_error(http_status, body):
            '''bail out when processing begun. Always HTTP 200 Ok.'''
            response['Response Status'] = http_status
            response['Response Body'] = body
            send_response(200)

        len_ = self._check_encoding()
        lines = self.rfile.read(len_).decode('utf-8').split("\n")
        for index, to_delete in enumerate(lines):
            if index >= self.MAX_DELETES:
                return inline_error(
                    '413 Request entity too large',
                    'Maximum Bulk Deletes: %d per request' % self.MAX_DELETES,
                )
            to_delete = urllib.parse.unquote(to_delete.strip())
            assert to_delete[0] == '/'
            to_delete = to_delete[1:].split('/', maxsplit=1)
            if len(to_delete) < 2:
                return error("deleting containers is not supported")
            to_delete = to_delete[1]
            try:
                del self.server.data[to_delete]
                del self.server.metadata[to_delete]
            except KeyError:
                response['Number Not Found'] += 1
            else:
                response['Number Deleted'] += 1

        if not (response['Number Deleted'] or response['Number Not Found']):
            return inline_error('400 Bad Request', 'Invalid bulk delete.')
        send_response(200)


#: A list of the available mock request handlers with
#: corresponding storage urls
handler_list = [
    (S3CRequestHandler, 's3c://%(host)s:%(port)d/s3ql_test'),
    (S3C4RequestHandler, 's3c4://%(host)s:%(port)d/s3ql_test'),
    # Special syntax only for testing against mock server
    (BasicSwiftRequestHandler, 'swift://%(host)s:%(port)d/s3ql_test'),
    (CopySwiftRequestHandler, 'swift://%(host)s:%(port)d/s3ql_test'),
    (BulkDeleteSwiftRequestHandler, 'swift://%(host)s:%(port)d/s3ql_test'),
]
